"""
  A collection of very generic python utilities.
  -- kandasamy@cs.cmu.edu
"""
from __future__ import print_function
from __future__ import division

# pylint: disable=import-error
# pylint: disable=no-member
# pylint: disable=no-name-in-module
# pylint: disable=invalid-name
# pylint: disable=relative-import

import numpy as np
from scipy.sparse import dok_matrix
from scipy.linalg import solve_triangular


# Some very generic utility functions
def map_to_cube(pts, bounds):
  """ Maps bounds to [0,1]^d and returns the representation in the cube. """
  return (pts - bounds[:, 0])/(bounds[:, 1] - bounds[:, 0])


def map_to_bounds(pts, bounds):
  """ Given a point in [0,1]^d, returns the representation in the original space. """
  return pts * (bounds[:, 1] - bounds[:, 0]) + bounds[:, 0]


def get_sublist_from_indices(orig_list, idxs):
  """ Returns a sublist from the indices. orig_list can be anthing that can be
      indexed and idxs are a list of indices. """
  return [orig_list[idx] for idx in idxs]


def project_to_bounds(x, bounds):
  """ Projects to bounds. """
  bounds = np.array(bounds)
  ret = np.clip(x, bounds[:, 0], bounds[:, 1])
  return ret


def get_idxs_from_list_of_lists(list_of_lists, idx):
  """ Returns a list of objects corresponding to index idx from a list of lists. """
  return [elem[idx] for elem in list_of_lists]


def transpose_list_of_lists(list_of_lists):
  """ Transposes a list of lists. """
  return [list(i) for i in zip(*list_of_lists)]


def compute_average_sq_prediction_error(Y1, Y2):
  """ Returns the average prediction error. """
  return np.linalg.norm(np.array(Y1) - np.array(Y2))**2 / len(Y1)


def dist_squared(X1, X2):
  """ If X1 is n1xd and X2 is n2xd, this returns an n1xn2 matrix where the (i,j)th
      entry is the squared distance between X1(i,:) and X2(j,:).
  """
  n1, dim1 = X1.shape
  n2, dim2 = X2.shape
  if dim1 != dim2:
    raise ValueError('Second dimension of X1 and X2 should be equal.')
  dist_sq = (np.outer(np.ones(n1), (X2**2).sum(axis=1))
             + np.outer((X1**2).sum(axis=1), np.ones(n2))
             - 2*X1.dot(X2.T))
  dist_sq = np.clip(dist_sq, 0.0, np.inf)
  return dist_sq


def pareto_dominates(u, v, tolerance=None):
  """ Returns true if u >= v elementwise, and at least
      one of the elements is not an equality.
  """
  u = np.asarray(u, dtype=np.float32)
  v = np.asarray(v, dtype=np.float32)
  if tolerance is None:
    return np.all(u >= v) and not np.all(u == v)
  else:
    return np.all(u >= v) and not np.linalg.norm(u - v) <= tolerance


def update_pareto_set(vals, points, new_val, new_point):
  """ Updates the current pareto optimal values and points with the
      new value and new point. vals is assummed to be a non-dominated
      set. The returned updated vals and points is guaranteed to
      be non-dominated provided the original values were
      non-dominated.
  """
  num_points = len(points)
  new_vals = []
  new_points = []
  for i in range(num_points):
    # If new_val does not dominate vals[i], keep it
    if not pareto_dominates(new_val, vals[i]):
      new_vals.append(vals[i])
      new_points.append(points[i])
  # Check if new_vals is not dominated by any vals[i].
  # If so, then keep it.
  dominated = False
  for i in range(num_points):
    if pareto_dominates(vals[i], new_val):
      dominated = True
      break
  if not dominated:
    new_points.append(new_point)
    new_vals.append(new_val)
  return new_vals, new_points


def pairwise_hamming_kernel(X1, X2, weights=None):
  """ Computes the pairwise hamming kernels between X1 and X2.
  """
  # An internal function to compute weighted distances between an array and a vector.
  def _compute_dist_between_2D_1D_array(arr_2d, arr_1d, wts):
    """ Returns distance between 2D and 1D array. """
    return (np.equal(arr_2d, arr_1d) * wts).sum(axis=1)
  # An internal function to compute weighted distances between a large array and
  # a small array
  def _compute_dist_between_large_small_array(arr_large, arr_small, wts):
    """ Returns distances between large and small arrays. """
    ret = np.zeros((arr_large.shape[0], arr_small.shape[0]))
    for idx, small_vec in enumerate(arr_small):
      ret[:, idx] = _compute_dist_between_2D_1D_array(arr_large, small_vec, wts)
    return ret
  # ---------------------------------------------------------------------------------
  if len(X1) == 0 and len(X2) == 0:
    return np.zeros((0, 0))
  elif len(X1) == 0:
    return np.zeros((0, len(X2)))
  elif len(X2) == 0:
    return np.zeros((len(X1), 0))
  else:
    if weights is None:
      dim = len(X1[0])
      weights = np.ones((dim, ))/float(dim)
    X1 = np.array(X1, dtype=np.object)
    X2 = np.array(X2, dtype=np.object)
    n1, _ = X1.shape
    n2, _ = X2.shape
    if n2 < n1:
      return _compute_dist_between_large_small_array(X2, X1, weights).T
    else:
      return _compute_dist_between_large_small_array(X1, X2, weights)


# Some linear algebraic utilities
def project_symmetric_to_psd_cone(M, is_symmetric=True, epsilon=0):
  """ Projects the symmetric matrix M to the PSD cone. """
  if is_symmetric:
    try:
      eigvals, eigvecs = np.linalg.eigh(M)
    except np.linalg.LinAlgError:
      print('LinAlgError encountered with eigh. Defaulting to eig.')
      eigvals, eigvecs = np.linalg.eig(M)
      eigvals = np.real(eigvals)
      eigvecs = np.real(eigvecs)
  else:
    eigvals, eigvecs = np.linalg.eig(M)
  clipped_eigvals = np.clip(eigvals, epsilon, np.inf)
  return (eigvecs * clipped_eigvals).dot(eigvecs.T)


def stable_cholesky(M, add_to_diag_till_psd=True):
  """ Returns L, a 'stable' cholesky decomposition of M. L is lower triangular and
      satisfies L*L' = M.
      Sometimes nominally psd matrices are not psd due to numerical issues. By adding a
      small value to the diagonal we can make it psd. This is what this function does.
      Use this iff you know that K should be psd. We do not check for errors.
  """
  if M.size == 0:
    return M # if you pass an empty array then just return it.
  try:
    # First try taking the Cholesky decomposition.
    L = np.linalg.cholesky(M)
  except np.linalg.linalg.LinAlgError as e:
    # If it doesn't work, then try adding diagonal noise.
    if not add_to_diag_till_psd:
      raise e
    diag_noise_power = -11
    max_M = np.diag(M).max()
    diag_noise = np.diag(M).max() * 1e-11
    chol_decomp_succ = False
    while not chol_decomp_succ:
      try:
        L = np.linalg.cholesky(M +
            ((10**diag_noise_power) * max_M)  * np.eye(M.shape[0]))
        chol_decomp_succ = True
      except np.linalg.linalg.LinAlgError:
        if diag_noise_power > -9:
          print('stable_cholesky failed with diag_noise_power=%d.'%(diag_noise_power))
        diag_noise_power += 1
      if diag_noise_power >= 5:
        print('**************** Cholesky failed: Added diag noise = %e'%(diag_noise))
  return L


# Solving triangular matrices -----------------------
def _solve_triangular_common(A, b, lower):
  """ Solves Ax=b when A is a triangular matrix. """
  if A.size == 0 and b.shape[0] == 0:
    return np.zeros((b.shape))
  else:
    return solve_triangular(A, b, lower=lower)

def solve_lower_triangular(A, b):
  """ Solves Ax=b when A is lower triangular. """
  return _solve_triangular_common(A, b, lower=True)

def solve_upper_triangular(A, b):
  """ Solves Ax=b when A is upper triangular. """
  return _solve_triangular_common(A, b, lower=False)


def draw_gaussian_samples(num_samples, mu, K):
  """ Draws num_samples samples from a Gaussian distribution with mean mu and
      covariance K.
  """
  num_pts = len(mu)
  L = stable_cholesky(K)
  U = np.random.normal(size=(num_pts, num_samples))
  V = L.dot(U).T + mu
  return V


# Executing a Pythong string -------------------------------------------------------
def evaluate_strings_with_given_variables(strs_to_execute, variable_dict=None):
  """ Executes a list of python strings and returns the results as a list.
      variable_dict is a dictionary mapping strings to values which may be used in
      str_to_execute.
  """
  if variable_dict is None:
    variable_dict = {}
  if not isinstance(strs_to_execute, list):
    got_list_of_constraints = False
    strs_to_execute = [strs_to_execute]
  else:
    got_list_of_constraints = True
  # Now set local variables
  for key, value in variable_dict.items():
    locals()[key] = value
  ret = []
  for elem in strs_to_execute:
    curr_result = eval(elem)
    ret.append(curr_result)
  if got_list_of_constraints:
    return ret
  else:
    return ret[0]

# Matrix/Array/List utilities ------------------------------------------------------
def get_nonzero_indices_in_vector(vec):
  """ Returns the nonzero indices in the vector vec. """
  if not isinstance(vec, np.ndarray):
    vec = np.asarray(vec.todense()).ravel()
  ret, = vec.nonzero()
  return ret

def reorder_list_or_array(M, ordering):
  """ Reorders a list or array like object. """
  if isinstance(M, list):
    return [M[i] for i in ordering]
  else:
    return M[ordering]

def reorder_list(L, ordering):
  """ reorders a list. """
  return reorder_list_or_array(L, ordering)

def get_original_order_from_reordered_list(L, ordering):
  """ Returns the original order from a reordered list. """
  ret = [None] * len(ordering)
  for orig_idx, ordered_idx in enumerate(ordering):
    ret[ordered_idx] = L[orig_idx]
  return ret

def flatten_nested_lists(L):
  """ Each element of L could be a list or any object other than a list. This function
      returns a list of non-list objects, where the internal lists or lists_of_lists
      have been 'flattened'.
  """
  ret = []
  for elem in L:
    if isinstance(elem, list):
      ret.extend(flatten_nested_lists(elem))
    else:
      ret.append(elem)
  return ret

def flatten_list_of_objects_and_lists(L):
  """ Each element of L could be a non-list object or a list of non-objects. This function
      returns a list of non-list objects, where the internal lists have been 'flattened'.
  """
  ret = []
  for elem in L:
    if isinstance(elem, list):
      ret.extend(elem)
    else:
      ret.append(elem)
  return ret

def flatten_list_of_objects_and_iterables(L):
  """ Each element of L could be a non-list object or a list of non-objects. This function
      returns a list of non-list objects, where the internal lists have been 'flattened'.
  """
  ret = []
  for elem in L:
    if hasattr(elem, '__iter__') and not isinstance(elem, str):
      ret.extend(elem)
    else:
      ret.append(elem)
  return ret

def flatten_list_of_lists(L):
  """ Flattens a list of lists to return a single list of objects. """
  return [item for sublist in L for item in sublist]

def reorder_rows_and_cols_in_matrix(M, ordering):
  """ Reorders the rows and columns in matrix M. """
  array_type = type(M)
  if array_type == dok_matrix: # Check if a sparse matrix to convert to array
    M = np.asarray(M.todense())
  elif array_type == list:
    M = np.array(M)
  # Now do the reordering
  M = M[:, ordering][ordering]
  # Convert back
  if array_type == dok_matrix: # Check if a sparse matrix for return
    M = dok_matrix(M)
  elif array_type == list:
    M = [list(m) for m in M]
  return M

def _set_coords_to_val(A, coords, val):
  """ Sets the indices in matrix A to value. """
  for coord in coords:
    A[coord[0], coord[1]] = val

def get_dok_mat_with_set_coords(n, coords):
  """ Returns a sparse 0 matrix with the coordinates in coords set to 1. """
  A = dok_matrix((n, n))
  _set_coords_to_val(A, coords, 1)
  return A

def block_augment_array(A, B, C, D):
  """ Given a n1xn2 array A, an n1xn3 array B, an n4xn5 array C, and a
      n4x(n2 + n3 - n5) array D, this returns (n1+n4)x(n2+n3) array of the form
      [A, B; C; D].
  """
  AB = np.hstack((A, B))
  CD = np.hstack((C, D))
  return np.vstack((AB, CD))


# For sampling based on fitness values -------------------------------------------------
# We are using them in the GA and BO algorithms.
def get_exp_probs_from_fitness(fitness_vals, scaling_param=None, scaling_const=None):
  """ Returns sampling probabilities from fitness values; the fitness values are
      exponentiated and used as probabilities.
  """
  fitness_vals = np.array(fitness_vals)
  if scaling_param is None:
    scaling_const = scaling_const if scaling_const is not None else 0.5
    scaling_param = scaling_const * fitness_vals.std()
  mean_param = fitness_vals.mean()
  exp_probs = np.exp((fitness_vals - mean_param)/scaling_param)
  return exp_probs/exp_probs.sum()

def sample_according_to_exp_probs(fitness_vals, num_samples, replace=False,
                                  scaling_param=None, scaling_const=None,
                                  sample_uniformly_if_fail=False):
  """ Samples after exponentiating the fitness values. """
  exp_probs = get_exp_probs_from_fitness(fitness_vals, scaling_param, scaling_const)
  if (not np.isfinite(exp_probs.sum())) and sample_uniformly_if_fail:
    exp_probs = np.ones((len(fitness_vals),)) / float(len(fitness_vals))
  return np.random.choice(len(fitness_vals), num_samples, p=exp_probs, replace=replace)




"""
  Unit tests for optimers.
  -- kandasamy@cs.cmu.edu
"""
from __future__ import absolute_import

# pylint: disable=import-error
# pylint: disable=no-member
# pylint: disable=invalid-name
# pylint: disable=relative-import

import os
import time
from argparse import Namespace
import numpy as np
# Local
from .ancillary_utils import is_nondecreasing
from .base_test_class import BaseTestClass, execute_tests
from .euclidean_synthetic_functions import get_mf_hartmann_function_data
from . import oper_utils

# TODO: Write unit tests for Optimal Transport


class OptimisersTestCase(BaseTestClass):
  """Unit test class for general utilities. """

  def __init__(self, *args, **kwargs):
    super(OptimisersTestCase, self).__init__(*args, **kwargs)

  def setUp(self):
    """ Sets up attributes. """
    self.problems = []
    self.max_evals = 20000 #1e4
    # First problem
    obj = lambda x: np.dot(x-1, x)
    dim = 4
    min_pt = np.array([0.5] * dim)
    max_pt = np.array([-1] * dim)
    self.problems.append(self._get_test_case_problem_instance(obj, [-1] * dim, \
                                               [1] * dim, min_pt, max_pt, '4D-quadratic'))
    # Second problem
    obj = lambda x: np.exp(-np.dot(x-0.5, x))
    dim = 2
    min_pt = np.array([-1] * dim)
    max_pt = np.array([0.25] * dim)
    self.problems.append(self._get_test_case_problem_instance(obj, [-1] * dim, \
                                                [1] * dim, min_pt, max_pt, '2D-gaussian'))
    # Third problem - use Hartmann6
    hartmann_data = get_mf_hartmann_function_data(0, 6)
    obj = hartmann_data[1]
    max_pt = hartmann_data[2]
    min_pt = None
    domain_bounds = np.array(hartmann_data[-1])
    self.problems.append(self._get_test_case_problem_instance(obj, domain_bounds[:, 0],
      domain_bounds[:, 1], min_pt, max_pt, 'Hartmann6'))

  @classmethod
  def _get_test_case_problem_instance(cls, obj, lb, ub, min_pt, max_pt, descr=''):
    """ A wrapper which returns a problem instance as a list. """
    min_val = float(obj(min_pt)) if min_pt is not None else None
    max_val = float(obj(max_pt)) if max_pt is not None else None
    bounds = np.vstack((lb, ub)).T
    problem_inst = Namespace(obj=obj, dim=len(lb), lb=lb, ub=ub, bounds=bounds,
                             min_pt=min_pt, max_pt=max_pt,
                             min_val=min_val, max_val=max_val,
                             descr=descr)
    return problem_inst

  def _test_optimise_method(self, minimise_method, maximise_method, test_success=True):
    """ Tests an optimiser method. """
    num_min_successes = 0
    num_max_successes = 0
    for prob in self.problems:
      # First the minimimum
      if prob.min_val is not None:
        min_val_soln, min_pt_soln, _ = minimise_method(prob.obj, prob.bounds,
                                                       self.max_evals)
        val_diff = abs(prob.min_val - min_val_soln)
        point_diff = np.linalg.norm(prob.min_pt - min_pt_soln)
        self.report(prob.descr +
          '(min):: true-val: %0.4f, soln: %0.4f, diff: %0.4f.'%(prob.min_val,
             min_val_soln, val_diff), 'test_result')
        self.report(prob.descr +
          '(min):: true-pt: %s, soln: %s, diff: %0.4f.'%(prob.min_pt, min_pt_soln,
             point_diff), 'test_result')
        min_is_successful = val_diff < 1e-3 and point_diff < 1e-3 * prob.dim
        num_min_successes += min_is_successful
      else:
        num_min_successes += 1
      # Now the maximum
      if prob.max_val is not None:
        max_val_soln, max_pt_soln, _ = maximise_method(prob.obj, prob.bounds,
                                                       self.max_evals)
        val_diff = abs(prob.max_val - max_val_soln)
        point_diff = np.linalg.norm(prob.max_pt - max_pt_soln)
        self.report(prob.descr +
          '(max):: true-val: %0.4f, soln: %0.4f, diff: %0.4f.'%(prob.max_val,
             max_val_soln, val_diff), 'test_result')
        self.report(prob.descr +
          '(max):: true-pt: %s, soln: %s, diff: %0.4f.'%(prob.max_pt, max_pt_soln,
             point_diff), 'test_result')
        max_is_successful = val_diff < 1e-3 and point_diff < 1e-3 * prob.dim
        num_max_successes += max_is_successful
      else:
        num_max_successes += max_is_successful
    # Check if successful
    if test_success:
      assert num_min_successes == len(self.problems)
      assert num_max_successes == len(self.problems)

  def test_random_optimise(self):
    """ Test direct optmisation."""
    self.report('Rand optimise:')
    random_minimise = lambda obj, bounds, max_evals: oper_utils.random_minimise(
                                                obj, bounds, max_evals, False, False)
    random_maximise = lambda obj, bounds, max_evals: oper_utils.random_maximise(
                                                obj, bounds, max_evals, False, False)
    self._test_optimise_method(random_minimise, random_maximise, False)

  def test_direct(self):
    """ Test direct optmisation."""
    self.report('DiRect minimise and maximise:')
    self._test_optimise_method(oper_utils.direct_ft_minimise,
                               oper_utils.direct_ft_maximise)

  def test_direct_times(self):
    """ Tests the running time of the package with and without file writing. """
    self.report('DiRect running times')
    log_file_names = ['', 'test_log']
    for prob in self.problems:
      clock_times = []
      real_times = []
      for log_file_name in log_file_names:
        start_clock = time.clock()
        start_real_time = time.time()
        _, _, _ = oper_utils.direct_ft_maximise(prob.obj, prob.bounds, self.max_evals,
                                                log_file_name=log_file_name)
        clock_time = time.clock() - start_clock
        real_time = time.time() - start_real_time
        clock_times.append(clock_time)
        real_times.append(real_time)
        if log_file_name:
          try:
            os.remove(log_file_name)
          except OSError:
            pass
      # Print results out
      result_str = ', '.join(['file: \'%s\': clk=%0.4f, real=%0.4f, #evals=%d'%(
        log_file_names[i], clock_times[i], real_times[i], self.max_evals) \
        for i in range(len(log_file_names))])
      self.report('%s:: %s'%(prob.descr, result_str), 'test_result')

  def test_direct_with_history(self):
    """ Tests direct with history. """
    self.report('DiRect with history.')
    for prob in self.problems:
      min_val_soln, _, history = oper_utils.direct_ft_maximise(prob.obj, prob.bounds, \
                                                      self.max_evals, return_history=True)
      if history is not None:
        assert is_nondecreasing(history.curr_opt_vals)
        assert np.abs(min_val_soln - history.curr_opt_vals[-1]) < 1e-4

  def test_pdoo(self):
    """ Test PDOO optmisation."""
    self.report('PDOO minimise and maximise:')
    self._test_optimise_method(oper_utils.pdoo_minimise, oper_utils.pdoo_maximise)

  def test_pdoo_times(self):
    """ Tests the running time of the package with and without file writing. """
    self.report('PDOO running times.')
    for prob in self.problems:
      start_clock = time.clock()
      start_real_time = time.time()
      _, _, _ = oper_utils.pdoo_maximise(prob.obj, prob.bounds, self.max_evals)
      clock_time = time.clock() - start_clock
      real_time = time.time() - start_real_time
      self.report('clk=%0.4f, real=%0.4f, #evals=%d'%(
                  clock_time, real_time, self.max_evals), 'test_result')


class SamplersTestCase(BaseTestClass):
  """Unit test class for sampling utilities. """

  def setUp(self):
    """ Sets up unit tests. """
    self.lhs_data = [(1, 10), (2, 5), (4, 10), (10, 100)]

  @classmethod
  def _check_sample_sizes(cls, data, samples):
    """ Data is a tuple of the form (dim, num_samples) ans samples is an ndarray."""
    assert (data[1], data[0]) == samples.shape

  def test_latin_hc_indices(self):
    """ Tests latin hyper-cube index generation. """
    self.report('Test Latin hyper-cube indexing. Only a sufficient condition check.')
    for data in self.lhs_data:
      lhs_true_sum = data[1] * (data[1] - 1) / 2
      lhs_idxs = oper_utils.latin_hc_indices(data[0], data[1])
      lhs_idx_sums = np.array(lhs_idxs).sum(axis=0)
      assert np.all(lhs_true_sum == lhs_idx_sums)

  def test_latin_hc_sampling(self):
    """ Tests latin hyper-cube sampling. """
    self.report('Test Latin hyper-cube sampling. Only a sufficient condition check.')
    for data in self.lhs_data:
      lhs_max_sum = float(data[1] + 1)/2
      lhs_min_sum = float(data[1] - 1)/2
      lhs_samples = oper_utils.latin_hc_sampling(data[0], data[1])
      lhs_sample_sums = lhs_samples.sum(axis=0)
      self._check_sample_sizes(data, lhs_samples)
      assert lhs_sample_sums.max() <= lhs_max_sum
      assert lhs_sample_sums.min() >= lhs_min_sum



if __name__ == '__main__':
  execute_tests()




"""
  Implements a base class for unit tests with some common utilities.
"""
from __future__ import print_function

import numpy as np
import random
import sys
from time import time
import unittest


class BaseTestClass(unittest.TestCase):
  """ An abstract base class for unit tests. """

  def __init__(self, *args, **kwargs):
    """ Constructor. """
    super(BaseTestClass, self).__init__(*args, **kwargs)

  @classmethod
  def report(cls, msg, msg_type='test_header', out=sys.stdout):
    " Reports a message. """
    prefix = '' # default prefix and suffix
    suffix = '\n'
    if msg_type == 'header':
      suffix = '\n' + '=' * len(msg) + '\n'
    elif msg_type == 'test_header':
      prefix = ' * '
    elif msg_type == 'test_result':
      prefix = '     - '
    out.write(prefix + msg + suffix)
    out.flush()


def execute_tests(seed_val='time'):
  """ Executes the tests. """
  # Set seed value
  # pylint: disable=superfluous-parens
  # pylint: disable=no-member
  if seed_val is not None:
    if seed_val == 'time':
      seed_val = int(time()*10) % 100000
    elif not isinstance(seed_val, int):
      raise ValueError('seed_val should be \'time\', an integer or None.')
    print('Setting random seed to %d.'%(seed_val))
    np.random.seed(seed_val)
    random.seed(seed_val)
  # Run unit tests
  unittest.main()




"""
  Unit tests for random optimiser.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=abstract-class-little-used

import numpy as np
# Local imports
from . import random_optimiser as random_optimiser
from ..exd.worker_manager import SyntheticWorkerManager
from ..utils.ancillary_utils import is_nondecreasing, get_list_of_floats_as_str, \
                                  get_rounded_list
from ..utils.base_test_class import BaseTestClass, execute_tests
from ..utils.euclidean_synthetic_functions import get_syn_func_caller
from ..utils import reporters


# A base class for all Optimisers on Euclidean spaces ----------------------------
class EuclideanOptimisersBaseTestCase(object):
  """ A base test class for all optimisers on Euclidean domains. """
  # pylint: disable=no-member
  # pylint: disable=invalid-name

  def setUp(self):
    """ set up. """
    self.test_max_capital = 60 # for hartmann 6
    self.func_caller = get_syn_func_caller('hartmann6',
                                           noise_type='gauss', noise_scale=0.1)
    self.worker_manager_1 = SyntheticWorkerManager(1, time_distro='const')
    self.worker_manager_3 = SyntheticWorkerManager(3, time_distro='halfnormal')

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager, options, reporter):
    """ Instantiate a specific optimiser. """
    raise NotImplementedError('Implement in a child test class.')

  @classmethod
  def run_optimiser(cls, func_caller, worker_manager, max_capital, mode, *args, **kwargs):
    """ Runs optimiser. """
    raise NotImplementedError('Implement in a child test class.')

  def test_instantiation(self):
    """ Tests instantiation of the optimiser. """
    optimiser = self._child_instantiate_optimiser(self.func_caller, self.worker_manager_3,
                  options=None, reporter=reporters.get_reporter('silent'))
    self.report('Instantiated %s object.'%(type(optimiser)))
    for attr in dir(optimiser):
      if not attr.startswith('_'):
        self.report('optimiser.%s = %s'%(attr, str(getattr(optimiser, attr))),
                    'test_result')

  def _test_optimiser_results(self, opt_val, opt_point, history):
    """ Tests optimiser results. """
    assert opt_val == history.curr_opt_vals[-1]
    assert opt_point.shape[0] == self.func_caller.domain.dim
    assert is_nondecreasing(history.curr_opt_vals)
    assert is_nondecreasing(history.curr_true_opt_vals)
    self.report('True opt val sequence: %s.'%(
      get_list_of_floats_as_str(history.curr_true_opt_vals)))
    saved_in_history = [key for key, _ in list(history.__dict__.items()) if not
                        key.startswith('__')]
    self.report('Stored in history: %s.'%(saved_in_history), 'test_result')

  def test_optimisation_single(self):
    """ Test optimisation with a single worker. """
    self.report('')
    self.report('Testing %s with one worker.'%(type(self)))
    opt_val, opt_point, history = self.run_optimiser(self.func_caller,
      self.worker_manager_1, self.test_max_capital, 'asy')
    self._test_optimiser_results(opt_val, opt_point, history)
    self.report('')
    return opt_val, opt_point, history

  def test_optimisation_asynchronous(self):
    """ Testing random optimiser with three asynchronous workers. """
    self.report('Testing %s with three asynchronous workers.'%(type(self)))
    opt_val, opt_point, history = self.run_optimiser(self.func_caller,
      self.worker_manager_3, self.test_max_capital, 'asy')
    self._test_optimiser_results(opt_val, opt_point, history)
    self.report('')
    return opt_val, opt_point, history

  def test_optimisation_synchronous(self):
    """ Testing random optimiser with three synchronous workers. """
    self.report('Testing %s with three synchronous workers.'%(type(self)))
    opt_val, opt_point, history = self.run_optimiser(self.func_caller,
      self.worker_manager_3, self.test_max_capital, 'syn')
    self._test_optimiser_results(opt_val, opt_point, history)
    self.report('')
    return opt_val, opt_point, history


# A test class for MF random optimisers ---------------------------------------------
def get_multi_fidelity_history_str(history):
  """ Prints the history of the multifidelity optimisation. """
  fidel_dim = len(history.query_fidels[0])
  print_fidels = np.array([get_rounded_list(qf) for qf in history.query_fidels]).T
  # First the function values
  val_title_str = ', '.join(['fidel(dim=%d)'%(fidel_dim), 'val', 'curr_true_opt_val'])
  val_title_str = val_title_str + '\n' + '-' * len(val_title_str)
  print_vals = get_rounded_list(history.query_vals)
  print_curr_opt_vals = get_rounded_list(history.curr_true_opt_vals)
  val_history_mat_str_T = np.vstack((print_fidels, print_vals, print_curr_opt_vals))
  val_history_mat_list = [get_list_of_floats_as_str(row) for row in
                          val_history_mat_str_T.T]
  val_history_mat_str = '\n'.join(val_history_mat_list)
  # NOw the fidels
  fidel_title_str = ', '.join(['fidel(%d)'%(fidel_dim), 'eval_times', 'fidel_costs',
                                'cum(fidel_costs)', 'receive_times'])
  fidel_title_str = fidel_title_str + '\n' + '-' * len(fidel_title_str)
  fidel_history_mat_str_T = np.vstack((print_fidels, history.query_eval_times,
                    history.query_cost_at_fidels,
                    np.cumsum(history.query_cost_at_fidels), history.query_receive_times))
  fidel_history_mat_list = [get_list_of_floats_as_str(row) for row in
                            fidel_history_mat_str_T.T]
  fidel_history_mat_str = '\n'.join(fidel_history_mat_list)
  return val_title_str + '\n' + val_history_mat_str + '\n\n' + \
         fidel_title_str + '\n' + fidel_history_mat_str + '\n'


# Multi-fidelity Optimisers base test case
class MFEuclideanOptimisersBaseTestCase(EuclideanOptimisersBaseTestCase):
  """ A base test class for all optimisers on Euclidean domains. """
  # pylint: disable=abstract-method
  # pylint: disable=no-member

  def setUp(self):
    """ Set up. Override the set up from the parent class. """
    super(MFEuclideanOptimisersBaseTestCase, self).setUp()
    self.func_caller = get_syn_func_caller('hartmann6', noise_type='gauss',
                                           noise_scale=0.1, fidel_dim=1)

  def test_optimisation_single(self):
    """ Test optimisation with a single worker. """
    _, _, history = \
      super(MFEuclideanOptimisersBaseTestCase, self).test_optimisation_single()
    mf_history_str = get_multi_fidelity_history_str(history)
    self.report(mf_history_str)

  def test_optimisation_asynchronous(self):
    """ Testing random optimiser with four asynchronous workers. """
    _, _, history = \
      super(MFEuclideanOptimisersBaseTestCase, self).test_optimisation_asynchronous()
    mf_history_str = get_multi_fidelity_history_str(history)
    self.report(mf_history_str)


# Now the Testcases for Random Euclidean Optimisers =================================

class EuclideanRandomOptimiserTestCase(EuclideanOptimisersBaseTestCase, BaseTestClass):
# class EuclideanRandomOptimiserTestCase(EuclideanOptimisersBaseTestCase):
  """ Unittest for the EuclideanRandomOptimiser class. """

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager,
                                   options=None, reporter=None):
    """ Instantiates a random optimiser in Euclidean space. """
    return random_optimiser.EuclideanRandomOptimiser(func_caller, worker_manager, options,
                                                     reporter)

  @classmethod
  def run_optimiser(cls, func_caller, worker_manager, max_capital, mode, *args, **kwargs):
    """ Runs optimiser. """
    return random_optimiser.random_optimiser_from_func_caller(
             func_caller, worker_manager, max_capital, mode)


class MFEucRandomOptimiserTestCase(MFEuclideanOptimisersBaseTestCase, BaseTestClass):
# class MFEucRandomOptimiserTestCase(MFEuclideanOptimisersBaseTestCase):
  """ Unit tests for the Multi-fidelity Euclidean random optimiser. """

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager,
                                   options=None, reporter=None):
    """ Instantiates a random MF optimiser in Euclidean space. """
    return random_optimiser.MFEuclideanRandomOptimiser(func_caller, worker_manager,
                                                       options=options, reporter=reporter)

  @classmethod
  def run_optimiser(cls, func_caller, worker_manager, max_capital, mode, *args, **kwargs):
    """ Runs optimiser. """
    return random_optimiser.mf_random_optimiser_from_func_caller(
                              func_caller, worker_manager, max_capital, mode,
                              call_fidel_to_opt_prob=0.333)


if __name__ == '__main__':
  execute_tests()




"""
  Unit tests for cp_domain_utils.py
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=no-member

import os
import numpy as np
# Local imports
from . import cp_domain_utils
from ..test_data.park1_3.park1_3 import park1_3
from ..test_data.park1_3.park1_3_mf import park1_3_mf
from ..test_data.park2_4.park2_4 import park2_4
from ..test_data.park2_4.park2_4_mf import park2_4_mf
# from ..test_data.syn_cnn_2.syn_cnn_2 import syn_cnn_2
from ..utils.base_test_class import BaseTestClass, execute_tests
from ..utils.general_utils import get_idxs_from_list_of_lists


class CPDomainUtilsTestCase(BaseTestClass):
  """ A unit test to test ancillary utils in setting up a cartesian product GP. """

  def setUp(self):
    """ Set up. """
    file_dir = os.path.dirname(os.path.realpath(__file__))
    test_data_pardir = os.path.dirname(file_dir)
    self.domain_config_files = [
      test_data_pardir + '/test_data/example_configs/eg01.json',
      test_data_pardir + '/test_data/example_configs/eg02.json',
      test_data_pardir + '/test_data/example_configs/eg03.json',
      test_data_pardir + '/test_data/example_configs/eg01.pb',
      test_data_pardir + '/test_data/example_configs/eg02.pb',
      test_data_pardir + '/test_data/example_configs/eg04.pb',
      test_data_pardir + '/test_data/park2_4/config.json',
      test_data_pardir + '/test_data/park1_3/config.json',
#       test_data_pardir + '/test_data/syn_cnn_2/config.json',
      ]
    self.synthetic_functions = [
      (test_data_pardir + '/test_data/park2_4/config.json', park2_4),
      (test_data_pardir + '/test_data/park1_3/config.json', park1_3),
#       ('test_data/syn_cnn_2/config.json', syn_cnn_2),
      ]
    self.domain_fidel_synthetic_functions = [
      (test_data_pardir + '/test_data/park2_4/config_mf.json', park2_4_mf),
      (test_data_pardir + '/test_data/park1_3/config_mf.json', park1_3_mf),
      ]

  def test_load_domain(self):
    """ Test load domain. """
    # TODO: this only tests for function, not correctness.
    self.report('Testing loading domain from config files - tests only for correctness.')
    for idx, dcf in enumerate(self.domain_config_files):
      self.report('[%d/%d] Testing loading from %s.'%(
                  idx + 1, len(self.domain_config_files), dcf))
      domain, orderings = cp_domain_utils.load_cp_domain_from_config_file(dcf)
      self.report('Domain: %s'%(domain), 'test_result')
      self.report('Ordering: %s'%(orderings), 'test_result')
    self.report('', '')

  def test_load_config_file_without_fidel_space(self):
    """ Tests loading a configuration file - both domain and fidelity space. """
    self.report('Testing loading configuration file with no fidelity space.')
    for dcf in self.domain_config_files:
      configs = cp_domain_utils.load_config_file(dcf)
      assert not hasattr(configs, 'fidel_space')
    self.report('', '')

  @classmethod
  def _test_if_two_elements_are_equal(cls, l1, l2):
    """ Returns True if two lists of objects are equal. """
    for idx in range(len(l1)):
      l1_elem = l1[idx]
      l2_elem = l2[idx]
      assert (hasattr(l1_elem, '__iter__') and hasattr(l2_elem, '__iter__')) or \
             (not hasattr(l1_elem, '__iter__') and not hasattr(l2_elem, '__iter__'))
      if isinstance(l1_elem, np.ndarray) or isinstance(l2_elem, np.ndarray):
        assert np.linalg.norm(np.array(l1_elem) - np.array(l2_elem)) < 1e-3
      else:
        assert l1_elem == l2_elem

  @classmethod
  def _test_if_two_lists_of_objects_are_equal(cls, list_1, list_2):
    """ Returns True if two lists of objects are equal. """
    assert len(list_1) == len(list_2)
    for l1, l2 in zip(list_1, list_2):
      cls._test_if_two_elements_are_equal(l1, l2)

  def test_packing_and_unpacking_points(self):
    """ Tests packing and unpacking. """
    self.report('Testing packing and unpacking points.')
    num_samples = 8
    for idx, dcf in enumerate(self.domain_config_files):
      self.report('[%d/%d] Testing packing/unpacking from %s.'%(
                  idx + 1, len(self.domain_config_files), dcf))
      cp_dom, orderings = cp_domain_utils.load_cp_domain_from_config_file(dcf)
      self.report('Domain: %s'%(cp_dom), 'test_result')
      proc_samples_1 = cp_domain_utils.sample_from_cp_domain(cp_dom, num_samples)
      raw_samples_1 = [cp_domain_utils.get_raw_point_from_processed_point(x, cp_dom,
                         orderings.index_ordering, orderings.dim_ordering)
                       for x in proc_samples_1]
      proc_samples_2 = [cp_domain_utils.get_processed_point_from_raw_point(x, cp_dom,
                          orderings.index_ordering, orderings.dim_ordering)
                        for x in raw_samples_1]
      raw_samples_2 = [cp_domain_utils.get_raw_point_from_processed_point(x, cp_dom,
                         orderings.index_ordering, orderings.dim_ordering)
                       for x in proc_samples_2]
      self._test_if_two_lists_of_objects_are_equal(proc_samples_1, proc_samples_2)
      self._test_if_two_lists_of_objects_are_equal(raw_samples_1, raw_samples_2)
    self.report('', '')

  def test_function_loading_and_processing(self):
    """ Tests creation of the function. """
    self.report('Testing function evaluations before and after packing.')
    num_samples = 25
    for idx, (dcf, raw_func) in enumerate(self.synthetic_functions):
      cp_dom, orderings = cp_domain_utils.load_cp_domain_from_config_file(dcf)
      proc_func = cp_domain_utils.get_processed_func_from_raw_func_for_cp_domain(
                    raw_func, cp_dom, orderings.index_ordering, orderings.dim_ordering)
      # Draw samples
      proc_samples = cp_domain_utils.sample_from_cp_domain(cp_dom, num_samples)
      raw_samples = [cp_domain_utils.get_raw_point_from_processed_point(x, cp_dom,
                         orderings.index_ordering, orderings.dim_ordering)
                       for x in proc_samples]
      # Compute function values
      proc_vals = [proc_func(x) for x in proc_samples]
      raw_vals = [raw_func(x) for x in raw_samples]
      err = np.linalg.norm(np.array(proc_vals) - np.array(raw_vals))
      assert err < 1e-4 * np.linalg.norm(np.array(proc_vals))
      self.report('[%d/%d] Testing function evaluations for %s, Error=%0.5f.'%(
                  idx + 1, len(self.synthetic_functions), dcf, err), 'test_result')
    self.report('', '')

  # Tests for loading configuration files -------------------------------------------
  def test_load_config_file(self):
    """ Tests loading a configuration file - both domain and fidelity space. """
    self.report('Testing loading configuration file.')
    for idx, (dcf, _) in enumerate(self.domain_fidel_synthetic_functions):
      self.report('[%d/%d] Testing loading from %s.'%(
                  idx + 1, len(self.domain_fidel_synthetic_functions), dcf))
      config = cp_domain_utils.load_config_file(dcf)
      self.report('domain: %s'%(config.domain), 'test_result')
      self.report('domain_ordering: %s'%(config.domain_orderings), 'test_result')
      self.report('fidel_space: %s'%(config.fidel_space), 'test_result')
      self.report('fidel_to_opt: %s, (raw: %s)'%(
        config.fidel_to_opt, config.raw_fidel_to_opt), 'test_result')
      self.report('fidel_space_ordering: %s'%(config.fidel_space_orderings),
                  'test_result')
    self.report('', '')

  def test_packing_and_unpacking_with_config(self):
    """ Tests packing and unpacking of points in both fidelity and domain. """
    self.report('Testing packing and unpacking from a configuration file.')
    test_data = self.synthetic_functions[:4] + self.domain_fidel_synthetic_functions
    num_samples = 8
    for idx, (dcf, _) in enumerate(test_data):
      config = cp_domain_utils.load_config_file(dcf)
      has_fidel = hasattr(config, 'fidel_space')
      self.report('[%d/%d] Testing packing/unpacking from %s, has fidel = %d.'%(
        idx + 1, len(test_data), dcf, has_fidel))
      if has_fidel:
        self.report('fidel_space: %s'%(config.fidel_space), 'test_result')
      self.report('domain: %s'%(config.domain), 'test_result')
      proc_samples_1 = cp_domain_utils.sample_from_config_space(config, num_samples)
      raw_samples_1 = [cp_domain_utils.get_raw_from_processed_via_config(x, config)
                       for x in proc_samples_1]
      proc_samples_2 = [cp_domain_utils.get_processed_from_raw_via_config(x, config)
                        for x in raw_samples_1]
      raw_samples_2 = [cp_domain_utils.get_raw_from_processed_via_config(x, config)
                       for x in proc_samples_2]
      if has_fidel:
        fidel_proc_1 = get_idxs_from_list_of_lists(proc_samples_1, 0)
        fidel_proc_2 = get_idxs_from_list_of_lists(proc_samples_2, 0)
        fidel_raw_1 = get_idxs_from_list_of_lists(raw_samples_1, 0)
        fidel_raw_2 = get_idxs_from_list_of_lists(raw_samples_2, 0)
        domain_proc_1 = get_idxs_from_list_of_lists(proc_samples_1, 1)
        domain_proc_2 = get_idxs_from_list_of_lists(proc_samples_2, 1)
        domain_raw_1 = get_idxs_from_list_of_lists(raw_samples_1, 1)
        domain_raw_2 = get_idxs_from_list_of_lists(raw_samples_2, 1)
        # Test if equal
        self._test_if_two_lists_of_objects_are_equal(fidel_proc_1, fidel_proc_2)
        self._test_if_two_lists_of_objects_are_equal(fidel_raw_1, fidel_raw_2)
        self._test_if_two_lists_of_objects_are_equal(domain_proc_1, domain_proc_2)
        self._test_if_two_lists_of_objects_are_equal(domain_raw_1, domain_raw_2)
      else:
        self._test_if_two_lists_of_objects_are_equal(proc_samples_1, proc_samples_2)
        self._test_if_two_lists_of_objects_are_equal(raw_samples_1, raw_samples_2)
    self.report('', '')

  def test_multi_fidelity_function_loading_and_processing_with_config(self):
    """ Tests creation of the function. """
    self.report('Testing multi-fidelity function evals before/after packing with config.')
    num_samples = 25
    for idx, (dcf, raw_func) in enumerate(self.domain_fidel_synthetic_functions):
      config = cp_domain_utils.load_config_file(dcf)
      proc_func = cp_domain_utils.get_processed_func_from_raw_func_via_config(
                    raw_func, config)
      # Draw samples
      proc_samples_1 = cp_domain_utils.sample_from_config_space(config, num_samples)
      raw_samples_1 = [cp_domain_utils.get_raw_from_processed_via_config(x, config)
                       for x in proc_samples_1]
      proc_samples_2 = [cp_domain_utils.get_processed_from_raw_via_config(x, config)
                        for x in raw_samples_1]
      raw_samples_2 = [cp_domain_utils.get_raw_from_processed_via_config(x, config)
                       for x in proc_samples_2]
      # Compute function values
      proc_vals_1 = [proc_func(z, x) for (z, x) in proc_samples_1]
      proc_vals_2 = [proc_func(z, x) for (z, x) in proc_samples_2]
      raw_vals_1 = [raw_func(z, x) for (z, x) in raw_samples_1]
      raw_vals_2 = [raw_func(z, x) for (z, x) in raw_samples_2]
      err_1 = np.linalg.norm(np.array(proc_vals_1) - np.array(raw_vals_1))
      err_2 = np.linalg.norm(np.array(proc_vals_1) - np.array(proc_vals_2))
      err_3 = np.linalg.norm(np.array(proc_vals_1) - np.array(raw_vals_2))
      err = err_1 + err_2 + err_3
      assert err < 1e-4 * np.linalg.norm(np.array(proc_vals_1))
      self.report('[%d/%d] Testing multi-fidelity function evals for %s, Error=%0.5f.'%(
                  idx + 1, len(self.domain_fidel_synthetic_functions), dcf, err),
                  'test_result')
    self.report('', '')


if __name__ == '__main__':
  execute_tests()




"""
  Harness to evaluate methods on a given task and save results.
  -- kandasamy@cs.cmu.edu
"""

from argparse import Namespace
import random
from time import time
import pickle
from scipy.io import savemat as sio_savemat
import numpy as np
# Local imports
from .reporters import get_reporter

# Changes to variable names from previous version
# BasicExperimenter -> BaseMethodEvaluator
# experiment_name -> study_name
# num_experiments -> num_trials
# experiment_iter -> trial_iter
# run_experiments -> run_trials
# run_experiment_iteration -> run_trial_iteration

class BaseMethodEvaluator(object):
  """ Base class for evaluating multiple methods for an ED/Optimisation task. """

  def __init__(self, study_name, num_trials, save_file_name,
               save_file_extension='', reporter='default', random_seed_val='time'):
    """ Constructor.
        random_seed_val: If None we will not change the random seed. If it is
          'time' we will set it to a time based value. If it is an int, we will set
          the seed to that value.
    """
    self.study_name = study_name
    self.num_trials = num_trials
    if save_file_extension == '':
      save_file_parts = save_file_name.split('.')
      save_file_name = save_file_parts[0]
      save_file_extension = save_file_parts[1]
    self.save_file_extension = save_file_extension
    self.save_file_full_name = save_file_name + '.' + self.save_file_extension
    self.pickle_file_name = save_file_name + '.p'
    self.reporter = get_reporter(reporter)
    self.to_be_saved = Namespace(study_name=self.study_name)
    self.data_not_to_be_mat_saved = []
    self.data_not_to_be_pickled = []
    # We will need these going forward.
    self.trial_iter = 0
    # Set the random seed
    if random_seed_val is not None:
      if random_seed_val == 'time':
        random_seed_val = int(time() * 100) % 100000
      self.reporter.writeln('Setting random seed to %d.'%(random_seed_val))
      np.random.seed(random_seed_val)
      random.seed(random_seed_val)

  def save_results(self):
    """ Saves results in save_file_full_name. """
    self.reporter.write('Saving results (trial-iter:%d) to %s ...  '%(self.trial_iter, \
                         self.save_file_full_name))
    try:
      if self.save_file_extension == 'mat':
        dict_to_be_saved = vars(self.to_be_saved)
        dict_to_be_mat_saved = {key:val for key, val in dict_to_be_saved.items()
                                if key not in self.data_not_to_be_mat_saved}
        # Fix for crash in single fidelity case -- replacing None with 'xx'.
        for i in range(dict_to_be_mat_saved['query_eval_times'].shape[0]):
          if isinstance(dict_to_be_mat_saved['query_eval_times'][i, -1], str):
            continue
          query_eval_times = [val if val is not None else 'xx' for val in \
                              dict_to_be_mat_saved['query_eval_times'][i, -1]]
          dict_to_be_mat_saved['query_eval_times'][i, -1] = query_eval_times
        sio_savemat(self.save_file_full_name, mdict=dict_to_be_mat_saved)
      else:
        raise NotImplementedError('Only implemented saving mat files so far.')
      save_successful = True
    except IOError:
      save_successful = False
    # Report saving status
    if save_successful:
      self.reporter.writeln('successful.')
    else:
      self.reporter.writeln('unsuccessful!!')

  def save_pickle(self):
    """ Dumps to everything. """
    save_in = open(self.pickle_file_name, 'wb')
    dict_to_be_saved = vars(self.to_be_saved)
    dict_to_be_pickled = {key:val for key, val in dict_to_be_saved.items()
                          if key not in self.data_not_to_be_pickled}
    pickle.dump(dict_to_be_pickled, save_in)
    save_in.close()

  def terminate_now(self):
    """ Returns true if we should terminate now. Can be overridden in a child class. """
    return self.trial_iter >= self.num_trials

  def run_trials(self):
    """ This runs the trial. Each trial executes each method on a problem. """
    self.reporter.writeln(self.get_trial_header())
    while not self.terminate_now():
      # Prelims
      self.trial_iter += 1
      iter_header = ('\nEXP %d/%d:: '%(self.trial_iter, self.num_trials)
                     + self.get_iteration_header())
      iter_header += '\n' + '=' * len(iter_header) + '\n'
      self.reporter.writeln(iter_header)
      # R trial iteration.
      self.run_trial_iteration()
      # Save results
      self.save_results()
    # Wrap up the trials
    self.wrapup_trials()

  def get_trial_header(self):
    """ Something to pring before running all the trials. Can be overridden in a
        child class."""
    # pylint: disable=no-self-use
    return ''

  def get_iteration_header(self):
    """ A header for the particular iteration. """
    # pylint: disable=no-self-use
    return ''

  def run_trial_iteration(self):
    """ Implements the current iteration of the exeperiment. """
    raise NotImplementedError('Implement this in a child class.')

  def wrapup_trials(self):
    """ Any code to wrap up the trials goes here. """
    # pylint: disable=no-self-use
    pass




"""
  Implements various synthetic functions on NN architectures.
  -- kandasamy@cs.cmu.edu
"""
# pylint: disable=invalid-name

import numpy as np

def _get_vals_wo_None(iter_of_vals):
  """ Returns a list of values without Nones. """
  return [x for x in iter_of_vals if x is not None]

def _num_units_signal(num_units_vals, bias_val, decay):
  """ Signal on the number of units. """
  num_units_vals = np.array(_get_vals_wo_None(num_units_vals))
  return np.exp(-decay * abs(num_units_vals.mean() - bias_val))

def _degree_signal(in_degrees, out_degrees, bias_val, decay):
  """ Signal on the degrees. """
  avg_degree = (in_degrees.mean() + out_degrees.mean())/2.0
  return np.exp(-decay * abs(avg_degree - bias_val))

def _get_ip_op_distance_signal(ip_op_dist, bias_val, decay):
  """ Signal on distance from input to output. """
  return np.exp(-decay * abs(ip_op_dist - bias_val))

def _get_layer_degree_signal(degree_of_layer, bias_val, decay):
  """ A signal based on the degree of a layer. """
  return np.exp(-decay * abs(degree_of_layer - bias_val))

def _get_num_layers_signal(num_layers, bias_val, decay):
  """ A signal based on the number of layers. """
  return np.exp(-decay * abs(num_layers - bias_val))

def _get_num_edges_signal(num_edges, bias_val, decay):
  """ A signal based on the total number of edges. """
  return np.exp(-decay * abs(num_edges - bias_val))

def _get_stride_signal(strides, bias_val, decay):
  """ A signal using the strides. """
  strides = np.array(_get_vals_wo_None(strides))
  return np.exp(-decay * abs(strides.mean() - bias_val))

def _get_conv_signal(layer_labels):
  """ A signal using the convolutional layers. """
  conv_layers = [ll for ll in layer_labels if \
                 ll.startswith('conv') or ll.startswith('res')]
  conv_filter_vals = np.array([float(ll[-1]) for ll in conv_layers])
  return (conv_filter_vals == 3).sum() / float(len(conv_filter_vals) + 1)

def _get_sigmoid_signal(layer_labels):
  """ A function using the sigmoid layer fraction as the signal. """
  internal_layers = [ll for ll in layer_labels if ll not in ['ip', 'op', 'linear']]
  good_layers = [ll in ['logistic', 'relu'] for ll in internal_layers]
  return sum(good_layers) / float(len(internal_layers) + 1)


def syn_func1_common(nn):
  """ A synthetic function on NN architectures. """
  return _num_units_signal(nn.num_units_in_each_layer, 1000, 0.002) + \
         _degree_signal(nn.get_in_degrees(), nn.get_out_degrees(), 5, 0.4) + \
         _get_ip_op_distance_signal(nn.get_distances_from_ip()[nn.get_op_layer_idx()],
                                    10, 0.2) + \
         _get_layer_degree_signal(nn.get_in_degrees()[nn.get_op_layer_idx()], 3, 0.5) + \
         _get_layer_degree_signal(nn.get_out_degrees()[nn.get_ip_layer_idx()], 4, 0.5) + \
         _get_num_layers_signal(nn.num_layers, 30, 0.1) + \
         _get_num_edges_signal(nn.conn_mat.sum(), 100, 0.05)


def cnn_syn_func1(nn):
  """ A synthetic function for CNNs. """
  return syn_func1_common(nn) + \
         _num_units_signal(nn.num_units_in_each_layer, 500, 0.001) + \
         _get_num_layers_signal(nn.num_layers, 50, 0.3) + \
         _get_stride_signal(nn.strides, 1.5, 3.0) + \
         _get_conv_signal(nn.layer_labels)


def mlp_syn_func1(nn):
  """ A synthetic function for MLPs. """
  return syn_func1_common(nn) + \
         _get_num_edges_signal(nn.conn_mat.sum(), 50, 0.1) + \
         _num_units_signal(nn.num_units_in_each_layer, 2000, 0.001) + \
         _get_sigmoid_signal(nn.layer_labels)




"""
  A collection of utilities for ancillary purposes.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=import-error
# pylint: disable=no-member
# pylint: disable=invalid-name
# pylint: disable=relative-import

import numpy as np


def compare_dict(dict_1, dict_2):
  """ Compares two dictionaries. """
  # N.B: Taken from stackoverflow:
  # http://stackoverflow.com/questions/4527942/comparing-two-dictionaries-in-python
  dict_1_keys = set(dict_1.keys())
  dict_2_keys = set(dict_2.keys())
  intersect_keys = dict_1_keys.intersection(dict_2_keys)
  added = dict_1_keys - dict_2_keys
  removed = dict_2_keys - dict_1_keys
  modified = {o: (dict_1[o], dict_2[o]) for o in intersect_keys if dict_1[o] != dict_2[o]}
  same = set(o for o in intersect_keys if dict_1[o] == dict_2[o])
  return added, removed, modified, same


# Tests for monotonicity
def is_nondecreasing(arr):
  """ Returns true if the sequence is non-decreasing. """
  return all([x <= y for x, y in zip(arr, arr[1:])])

def is_nonincreasing(arr):
  """ Returns true if the sequence is non-increasing. """
  return all([x >= y for x, y in zip(arr, arr[1:])])

def is_increasing(arr):
  """ Returns true if the sequence is increasing. """
  return all([x < y for x, y in zip(arr, arr[1:])])

def is_decreasing(arr):
  """ Returns true if the sequence is decreasing. """
  return all([x > y for x, y in zip(arr, arr[1:])])

def dicts_are_equal(dict_1, dict_2):
  """ Returns true if dict_1 and dict_2 are equal. """
  added, removed, modified, _ = compare_dict(dict_1, dict_2)
  return len(added) == 0 and len(removed) == 0 and len(modified) == 0


# Print lists as strings
def get_rounded_list(float_list, round_to_decimals=3):
  """ Rounds the list and returns. """
  ret = np.array(float_list).round(round_to_decimals)
  if isinstance(float_list, list):
    ret = list(ret)
  return ret

def get_list_as_str(list_of_objs):
  """ Returns the list as a string. """
  return '[' + ' '.join([str(x) for x in list_of_objs]) + ']'

def get_list_of_floats_as_str(float_list, round_to_decimals=3):
  """ Rounds the list and returns a string representation. """
  float_list = get_rounded_list(float_list, round_to_decimals)
  return get_list_as_str(float_list)


# Some plotting utilities.
def plot_2d_function(func, bounds, x_label='x', y_label='y', title=None):
  """ Plots a 2D function in bounds. """
  # pylint: disable=unused-variable
  dim_grid_size = 20
  x_grid = np.linspace(bounds[0][0], bounds[0][1], dim_grid_size)
  y_grid = np.linspace(bounds[1][0], bounds[1][1], dim_grid_size)
  XX, YY = np.meshgrid(x_grid, y_grid)
  f_vals = func(XX.ravel(), YY.ravel())
  FF = f_vals.reshape(dim_grid_size, dim_grid_size)
  # Create plot
  import matplotlib.pyplot as plt
  from mpl_toolkits.mplot3d import Axes3D
  fig = plt.figure()
  ax = fig.gca(projection='3d')
  ax.plot_surface(XX, YY, FF)
  plt.xlabel(x_label)
  plt.ylabel(y_label)
  if title is not None:
    plt.title(title)
  return fig, ax, plt




"""
  Unit tests for posterior sampling
  -- kvysyara@andrew.cmu.edu
"""
from __future__ import absolute_import
from __future__ import division
import unittest

# pylint: disable=invalid-name
# pylint: disable=no-self-use

import numpy as np

# Local imports
from .continuous import Beta
from .model import Model
from ..utils.base_test_class import BaseTestClass, execute_tests

class PosteriorMixedTestCase(BaseTestClass):
  """ Unit test for posterior sampling with Bernoulli likelihood and Beta prior"""

  def setUp(self):
    """ Sets up unit tests. """
    self.size = 10000
    self.n = 1000
    self.threshold = 0.1
    self.alpha = 2
    self.beta = 3
    self.data = np.random.binomial(1, 0.6, self.n)
    self.Y = np.sum(self.data)
    self.prior = Beta(self.alpha, self.beta)

  def _check_sample_sizes(self, samples):
    """ Compares the sample sizes with the size parameter"""
    assert self.size == len(samples)

  def _compute_mean(self, samples):
    """ Computes Mean """
    return np.mean(samples)

  def _compute_variance(self, samples):
    """ Computes Variance """
    return np.var(samples)

  def _pdf(self, x):
    """ Computes pdf of model """
    if x < 0 or x > 1:
      return 0
    value = np.power(x, self.Y + self.alpha - 1)
    value = value*np.power(1 - x, self.n - self.Y + self.beta - 1)
    return value

  def _logp(self, x):
    """ Compute log pdf of model """
    if x < 0 or x > 1:
      return -np.inf
    value = (self.Y + self.alpha - 1)*np.log(x)
    value = value + (self.n - self.Y + self.beta - 1)*np.log(1 - x)
    return value

  def _grad_logp(self, x):
    """ Computes gradient of log pdf of model """
    if x < 0 or x > 1:
      return 0
    value = (self.Y + self.alpha - 1.0)/x
    value = value + (self.n - self.Y + self.beta - 1.0)/(1 - x)
    return value

  @unittest.skip
  def test_posterior_NUTS(self):
    """ Tests posterior estimation using NUTS sampling """
    self.report('Test posterior estimation with Bernoulli likelihood and Beta prior '
                'using NUTS sampling')

    model_a = Model(self._pdf, self._logp, self._grad_logp)
    samples = model_a.draw_samples('nuts', self.size, np.array([0.1]), 10)

    # Map estimation
    alpha_a = self.Y + self.alpha
    beta_a = self.n - self.Y + self.beta
    mean_a = alpha_a/float(alpha_a + beta_a)
    var_a = (alpha_a*beta_a)/((np.power(alpha_a+beta_a, 2))*(alpha_a+beta_a+1))

    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)

    self._check_sample_sizes(samples)
    assert abs(mean_a - mean_r) <= self.threshold
    assert abs(var_a - var_r) <= self.threshold
    self.report('Map Estimation: mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(mean_a, var_a, mean_r, var_r), 'test_result')

  def test_posterior_Slice(self):
    """ Tests posterior estimation using Slice sampling """
    self.report('Test posterior estimation with Bernoulli likelihood and Beta prior '
                'using Slice sampling')

    model_a = Model(self._pdf, self._logp, self._grad_logp)
    samples = model_a.draw_samples('slice', self.size, np.array([0.1]))

    # Map estimation
    alpha_a = self.Y + self.alpha
    beta_a = self.n - self.Y + self.beta
    mean_a = alpha_a/float(alpha_a + beta_a)
    var_a = (alpha_a*beta_a)/((np.power(alpha_a+beta_a, 2))*(alpha_a+beta_a+1))

    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)

    self._check_sample_sizes(samples)
    assert abs(mean_a - mean_r) <= self.threshold
    assert abs(var_a - var_r) <= self.threshold
    self.report('Map Estimation: mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(mean_a, var_a, mean_r, var_r), 'test_result')

if __name__ == '__main__':
  execute_tests()



"""
  A GA optimiser for Neural Networks.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name

from argparse import Namespace
import numpy as np
# Local imports
from .nn_opt_utils import get_initial_pool
from ..opt.ga_optimiser import GAOptimiser, ga_opt_args
from ..utils.option_handler import load_options
from ..utils.reporters import get_reporter


nn_ga_opt_args = ga_opt_args


class NNGAOptimiser(GAOptimiser):
  """ Optimising functions defined on Neural Networks. """
  # pylint: disable=abstract-method

  def _get_initial_qinfos(self, num_init_evals):
    """ Initial qinfos. """
    ret = get_initial_pool(self.domain.nn_type)
    np.random.shuffle(ret)
    ret = ret[:num_init_evals]
    return [Namespace(point=x) for x in ret]

# APIs
# ======================================================================================
def nn_ga_optimise_from_args(func_caller, worker_manager, max_capital, mode, mutation_op,
                             crossover_op=None, options=None, reporter='default'):
  """ GA optimisation from args. """
  if options is None:
    reporter = get_reporter(reporter)
    options = load_options(ga_opt_args, reporter=reporter)
  options.mode = mode
  return (NNGAOptimiser(func_caller, worker_manager, mutation_op, crossover_op,
                        options, reporter)).optimise(max_capital)




"""
  Unit tests for classes/methods in euclidean_gp.py
  -- kandasamy@cs.cmu.edu
"""
from __future__ import division
import unittest

# pylint: disable=invalid-name

import numpy as np
# Local
from .euclidean_gp import EuclideanGP, EuclideanGPFitter, euclidean_gp_args
from .unittest_gp_core import compute_average_prediction_error
from .unittest_gp_core import gen_gp_test_data
from ..utils.base_test_class import BaseTestClass, execute_tests
from ..utils.option_handler import load_options


# Some wrappers to build a GP ----------------------------------------------------------
def build_euc_gp_with_dataset(dataset, kernel_type):
  """ Function to build a SE GP with the dataset. """
  mean_func = lambda x: np.array([np.median(dataset[1])] * len(x))
  noise_var = dataset[1].std()**2/20
  kernel_hyperparams = dataset[2].hyperparams
  kernel_hyperparams['dim'] = len(dataset[0][0])
  kernel_hyperparams['nu'] = 2.5
  return EuclideanGP(dataset[0], dataset[1], kernel_type, mean_func, noise_var,
                     kernel_hyperparams)

def build_se_gp_with_dataset(dataset):
  """ Builds an SE GP with the dataset. """
  return build_euc_gp_with_dataset(dataset, 'se')

def build_matern_gp_with_dataset(dataset):
  """ builds an se gp with the dataset. """
  return build_euc_gp_with_dataset(dataset, 'matern')

def fit_se_gp_with_dataset(dataset):
  """ A wrapper to fit a gp using the dataset. """
  options = load_options(euclidean_gp_args)
  options.kernel_type = 'se'
  ret_fit_gp = (EuclideanGPFitter(dataset[0], dataset[1],
                                  options=options, reporter=None)).fit_gp()
  assert ret_fit_gp[0] == 'fitted_gp'
  return ret_fit_gp[1]

def fit_matern_gp_with_dataset(dataset, nu=-1.0):
  """ A wrapper to fit a GP with a matern kernel using the dataset. """
  options = load_options(euclidean_gp_args)
  options.kernel_type = 'matern'
  options.matern_nu = nu
  ret_fit_gp = (EuclideanGPFitter(dataset[0], dataset[1],
                                  options=options, reporter=None)).fit_gp()
  assert ret_fit_gp[0] == 'fitted_gp'
  return ret_fit_gp[1]

def fit_poly_gp_with_dataset(dataset, poly_order):
  """ A wrapper to fit a GP with a polynomial kernel using the dataset. """
  #pylint: disable=unused-argument
  raise NotImplementedError('Not implemented yet.')

def fit_add_gp_with_dataset(dataset, options):
  """ Fits an additive gp with dataset. """
  options.use_additive_gp = True
  ret_fit_gp = (EuclideanGPFitter(dataset[0], dataset[1],
                                  options=options, reporter=None)).fit_gp()
  assert ret_fit_gp[0] == 'fitted_gp'
  return ret_fit_gp[1]

def fit_add_se_gp_with_dataset(dataset):
  """ Fits an Additive GP with SE kernels. """
  options = load_options(euclidean_gp_args)
  options.kernel_type = 'se'
  return fit_add_gp_with_dataset(dataset, options)

def fit_add_matern_gp_with_dataset(dataset, nu=1.5):
  """ Fits an Additive GP with Matern kernels. """
  options = load_options(euclidean_gp_args)
  options.kernel_type = 'matern'
  options.matern_nu = nu
  return fit_add_gp_with_dataset(dataset, options)


# A base class that needs to be inherited by all classes ------------------------
class EuclideanGPFitterBaseTestClass(object):
  """ Some common functionality for all Euclidean tests. """
  # pylint: disable=no-member

  def setUp(self):
    """ Set up for the tests. """
    self.datasets = gen_gp_test_data()
    self.datasets = self.datasets[0:2]

  def _marg_likelihood_test(self, get_gp_func1, get_gp_func2, descr1, descr2):
    """ This tests for the marginal likelihood. GP2 is expected to do better. """
    num_successes = 0
    for dataset in self.datasets:
      gp1 = get_gp_func1(dataset)
      gp2 = get_gp_func2(dataset)
      lml1 = gp1.compute_log_marginal_likelihood()
      lml2 = gp2.compute_log_marginal_likelihood()
      success = lml1 <= lml2
      self.report('(N,D)=%s:: %s-lml=%0.4f, %s-lml=%0.4f, succ=%d'%(
          str(dataset[0].shape), descr1, lml1, descr2, lml2, success), 'test_result')
      num_successes += success
    assert num_successes == len(self.datasets)

  def _prediction_test(self, get_gp_func1, get_gp_func2, descr1, descr2):
    """ Tests for prediction on a test set. GP2 is expected to do better. """
    num_successes = 0
    for dataset in self.datasets:
      gp1 = get_gp_func1(dataset)
      preds1, _ = gp1.eval(dataset[3])
      err1 = compute_average_prediction_error(dataset, preds1)
      gp2 = get_gp_func2(dataset)
      preds2, _ = gp2.eval(dataset[3])
      err2 = compute_average_prediction_error(dataset, preds2)
      success = err2 <= err1
      self.report('(N,D)=%s:: %s-err=%0.4f, %s-err=%0.4f, succ=%d'%(
          str(dataset[0].shape), descr1, err1, descr2, err2, success), 'test_result')
      self.report('  -- GP: %s'%(str(gp2)), 'test_result')
      num_successes += success
    assert num_successes > 0.6 *len(self.datasets)


# Test for naive predictions vs fitted predictions --------------------------------
@unittest.skip
class EuclideanGPFitterTestCase(EuclideanGPFitterBaseTestClass, BaseTestClass):
  """ Unit tests for the EuclideanGP and EuclideanGPFitter class. """
  # pylint: disable=too-many-locals

  def test_se_marg_likelihood(self):
    """ This tests for the marginal likelihood for an SE GP. """
    self.report('Marginal likelihood for SEGP. Probabilistic test, might fail.')
    self._marg_likelihood_test(build_se_gp_with_dataset, fit_se_gp_with_dataset,
                               'naive', 'fitted')

  def test_se_prediction(self):
    """ Tests for prediction on a test set with an SEGP. """
    self.report('Prediction for an SE kernel. Probabilistic test, might fail.')
    self._prediction_test(build_se_gp_with_dataset, fit_se_gp_with_dataset,
                          'naive', 'fitted')

  def test_matern_marg_likelihood(self):
    """ This tests for the marginal likelihood for a Matern GP. """
    self.report('Marginal likelihood for MaternGP. Probabilistic test, might fail.')
    self._marg_likelihood_test(build_matern_gp_with_dataset, fit_matern_gp_with_dataset,
                               'naive', 'fitted')

  def test_matern_prediction(self):
    """ Tests for prediction on a test set with a MaternGP. """
    self.report('Prediction for a Matern kernel. Probabilistic test, might fail.')
    self._prediction_test(build_matern_gp_with_dataset, fit_matern_gp_with_dataset,
                          'naive', 'fitted')


# Test for non-additive vs additive predictions --------------------------------
@unittest.skip
class AdditiveGPFitterTestCase(EuclideanGPFitterBaseTestClass, BaseTestClass):
# class AdditiveGPFitterTestCase(EuclideanGPFitterBaseTestClass):
  """ Unit tests for the additive kernel. """

  def setUp(self):
    """ Set up for the tests. """
    self.datasets = gen_gp_test_data()
    self.datasets = self.datasets[1:3]

  def test_se_marg_likelihood(self):
    """ Test for the Marginal likelihood of an additive GP. """
    self.report('Marginal Likelihoods for non-additive vs additive SE GP.' +
                ' Probabilistic test, might fail.')
    self._marg_likelihood_test(fit_se_gp_with_dataset, fit_add_se_gp_with_dataset,
                               'non-additive', 'additive')

  def test_se_prediction(self):
    """ Test for the Marginal likelihood of an additive GP. """
    self.report('Marginal Likelihoods for non-additive vs additive SE GP.' +
                ' Probabilistic test, might fail.')
    self._prediction_test(fit_se_gp_with_dataset, fit_add_se_gp_with_dataset,
                          'non-additive', 'additive')

  def test_matern_marg_likelihood(self):
    """ Test for the Marginal likelihood of an additive GP. """
    self.report('Marginal Likelihoods for non-additive vs additive Matern GP.' +
                ' Probabilistic test, might fail.')
    self._marg_likelihood_test(fit_matern_gp_with_dataset, fit_add_matern_gp_with_dataset,
                               'non-additive', 'additive')

  def test_matern_prediction(self):
    """ Test for the Marginal likelihood of an additive GP. """
    self.report('Marginal Likelihoods for non-additive vs additive Matern GP.' +
                ' Probabilistic test, might fail.')
    self._prediction_test(fit_matern_gp_with_dataset, fit_add_matern_gp_with_dataset,
                          'non-additive', 'additive')



if __name__ == '__main__':
  execute_tests(2349)
#   execute_tests()




"""
  Parser for json and protocol buffer files
  -- kvysyara@andrew.cmu.edu
"""

# pylint: disable=invalid-name

from __future__ import absolute_import

import sys
import json
from collections import OrderedDict
from numbers import Number
import numpy as np

# Python 3 issue with unicode classes
try:
  UNICODE_EXISTS = bool(type(unicode))
except NameError:
  unicode = str


def unicode_to_str(data):
  """ Unicode to string conversion. """
  if not isinstance(data, str):
    return data.encode('utf-8')
  return data


def _load_fidel_to_opt_parameters(param):
  """ Loads fidel_to_opt parameters. """
  if isinstance(param, (list, tuple)):
    ret = []
    for elem in param:
      ret.append(_load_fidel_to_opt_parameters(elem))
  else:
    ret = param # just return the param
    if isinstance(param, unicode):
      ret = unicode_to_str(ret)
  return ret


def _load_domain_constraints(domain_constraints):
  """ Loads the domain constraints. """
  processed_constraints = []
  # The constraints will be represented as a list of 3-tuples. The first of each tuple
  # will be the name of the constraint, the second will be the constraint, and the third
  # to store any ancillary information. The third item will be a dictionary containing
  # any other information specified in the configuration file.
  for _, constraint_data in domain_constraints.items():
    curr_constraint_ancillary_info = {}
    for key, val in constraint_data.items():
      key = unicode_to_str(key)
      val = unicode_to_str(val)
      if key == 'name':
        curr_constraint_name = val
      elif key == 'constraint':
        curr_constraint = val
      else:
        curr_constraint_ancillary_info[key] = val
    curr_constraint_tuple = (curr_constraint_name, curr_constraint,
                             curr_constraint_ancillary_info)
    processed_constraints.append(curr_constraint_tuple)
    del curr_constraint_name
    del curr_constraint
    del curr_constraint_ancillary_info
  return processed_constraints



def load_parameters(config):
  """ Parses all the parameters from json config file. """
  # pylint: disable=too-many-branches
  exp_info = {}
  exp_info['name'] = unicode_to_str(config.get('name'))
  if exp_info['name'] is None:
    raise ValueError('Experiment name is required')
  # Domain -------------------------------------------------------
  parameters = []
  _parameters = config['domain']
  if isinstance(_parameters, dict):
    for key in list(_parameters.keys()):
      param = load_parameter(_parameters[key], key)
      parameters.append(param)
  elif isinstance(_parameters, list):
    for _parameter in _parameters:
      param = load_parameter(_parameter)
      parameters.append(param)
  else:
    raise ValueError('Wrong parameter type.')
  # domain_constraints -------------------------------------------
  domain_constraints = config.get('domain_constraints', None)
  if domain_constraints is not None:
    domain_constraints = _load_domain_constraints(domain_constraints)
  # Fidelity space -----------------------------------------------
  fidel_parameters = []
  _parameters = config.get('fidel_space', {})
  if isinstance(_parameters, dict):
    for key in list(_parameters.keys()):
      param = load_parameter(_parameters[key], key)
      fidel_parameters.append(param)
  elif isinstance(_parameters, list):
    for _parameter in _parameters:
      param = load_parameter(_parameter)
      fidel_parameters.append(param)
  else:
    raise ValueError('Wrong parameter type.')
  # fidel_space_constraints ---------------------------------------
  fidel_space_constraints = config.get('fidel_space_constraints', None)
  if fidel_space_constraints is not None:
    fidel_space_constraints = _load_domain_constraints(fidel_space_constraints)
  # fidel_to_opt --------------------------------------------------
  fidel_to_opt = config.get('fidel_to_opt', None)
  if fidel_to_opt is not None:
    fidel_to_opt = _load_fidel_to_opt_parameters(fidel_to_opt)
  # Return
  return {'exp_info':exp_info, 'domain':parameters, 'fidel_space':fidel_parameters,
          'fidel_to_opt':fidel_to_opt, 'domain_constraints':domain_constraints,
          'fidel_space_constraints':fidel_space_constraints}


def load_parameter(parameter, key=None):
  """ Parses each parameter and return a dict """
  # pylint: disable=too-many-branches
  # pylint: disable=too-many-statements
  # Common parameters
  _name = parameter.get('name', key)
  if _name is None:
    raise ValueError('Parameter name is required')
  _type = parameter.get('type', 'float')
  _kernel = parameter.get('kernel', '')
  # Common for Euclidean, Integral, Discrete, Discrete Numeric domains
  _dim = parameter.get('dim', "")
  # For Euclidean/Integral
  _min = parameter.get('min', -np.inf)
  _max = parameter.get('max', np.inf)
  # For Discrete
  _items = parameter.get('items', '')
  # For neural networks -- see below

  # Now process them
  param = {}
  param['name'] = unicode_to_str(_name)
  param['kernel'] = unicode_to_str(_kernel)
  param['type'] = unicode_to_str(_type).lower()
  # First for regular domains
  if param['type'] in ['float', 'int', 'discrete', 'discrete_numeric', 'boolean']:
    if not isinstance(_dim, Number):
      _dim = unicode_to_str(_dim)
    if _dim != "":
      _dim = int(_dim)
    param['dim'] = _dim
    if param['type'] in ['float', 'int']:
      param['min'] = _min
      param['max'] = _max
    elif param['type'] == 'discrete':
      if _items == '':
        raise ValueError('List of items required')
      param['items'] = unicode_to_str(_items).split('-')
    elif param['type'] == 'discrete_numeric':
      if _items == '':
        raise ValueError('List or range of items required')
      elif ':' not in _items:
        param['items'] = [float(x) for x in unicode_to_str(_items).split('-')]
      else:
        _range = [float(x) for x in unicode_to_str(_items).split(':')]
        param['items'] = list(np.arange(_range[0], _range[2], _range[1]))
  elif param['type'].startswith(('cnn', 'mlp')):
    nn_params = {}
    nn_params['max_num_layers'] = parameter.get('max_num_layers', 'inf')
    nn_params['min_num_layers'] = parameter.get('min_num_layers', 0)
    nn_params['max_mass'] = parameter.get('max_mass', 'inf')
    nn_params['min_mass'] = parameter.get('min_mass', 0)
    nn_params['max_in_degree'] = parameter.get('max_in_degree', 'inf')
    nn_params['max_out_degree'] = parameter.get('max_out_degree', 'inf')
    nn_params['max_num_edges'] = parameter.get('max_num_edges', 'inf')
    nn_params['max_num_units_per_layer'] = parameter.get('max_num_units_per_layer', 'inf')
    nn_params['min_num_units_per_layer'] = parameter.get('min_num_units_per_layer', 0)
    # For CNNs add strides
    if param['type'].startswith('cnn'):
      nn_params['max_num_2strides'] = parameter.get('max_num_2strides', 'inf')
    for nnp_key, nnp_val in nn_params.items():
      if isinstance(nnp_val, str):
        nnp_val = unicode_to_str(nnp_val)
      nnp_val = np.inf if nnp_val == 'inf' else nnp_val
      param[nnp_key] = nnp_val
    # Finally add the following
    param['dim'] = ''
  else:
    raise ValueError('Unknown type %s.'%(param['type']))
  return param


def read_json(config_file):
  """ Read from json file. """
  try:
    with open(config_file, 'r') as _file:
      config = json.load(_file, object_pairs_hook=OrderedDict)
      _file.close()
  except Exception as e:
    raise Exception('Error in loading config file: ' + config_file + '.\n -- ' + str(e))
  return load_parameters(config)


def read_pb(config_file):
  """ Read from protocol buffer file. """
  # pylint: disable=import-error
  try:
    from google.protobuf import text_format
    from google.protobuf.json_format import MessageToDict
  except ImportError:
    raise ImportError('Protocol Buffer library is not installed')
  # Read PB file
  from . import config_pb2
  config_pb = config_pb2.Experiment()
  _file = open(config_file, "rb")
  text_format.Merge(_file.read(), config_pb)
  _file.close()
  # Load parameters and return
  config = MessageToDict(config_pb)
  config['fidel_space'] = config.pop('fidelSpace', {})
  return load_parameters(config)


def config_parser(config_file):
  """Reads config files and creates domain objects. """
  if config_file.endswith('.json'):
    params = read_json(config_file)
  elif config_file.endswith('.pb'):
    params = read_pb(config_file)
  else:
    raise ValueError('Wrong Config file: %s' % (config_file))
  return params


if __name__ == '__main__':
  if len(sys.argv) < 2:
    raise ValueError('Need Config File.')
  config_parser(sys.argv[1])




#!/usr/bin/env python
import os
from setuptools import find_packages
from distutils.errors import CCompilerError, DistutilsExecError, DistutilsPlatformError, DistutilsError

from numpy.distutils.core import setup, Extension
from numpy.distutils.command.build_ext import build_ext as old_build_ext
from numpy.distutils.fcompiler import CompilerNotFound

class BuildFailed(Exception):
    pass


def construct_build_ext(build_ext):
    # This class allows extension building to fail.
    # https://stackoverflow.com/questions/41778153/
    ext_errors = (CCompilerError, DistutilsExecError,
                  DistutilsPlatformError, DistutilsError, IOError)
    class WrappedBuildExt(build_ext):
        def run(self):
            try:
                build_ext.run(self)
            except (DistutilsPlatformError, CompilerNotFound) as x:
                raise BuildFailed(x)

        def build_extension(self, ext):
            try:
                build_ext.build_extension(self, ext)
            except ext_errors as x:
                raise BuildFailed(x)
    return WrappedBuildExt


with open('README.md', 'r') as f:
    LONG_DESCRIPTION = f.read()


direct_files = ['direct.pyf', 'DIRect.f', 'DIRserial.f', 'DIRsubrout.f']
direct_paths = [os.path.join('dragonfly', 'utils', 'direct_fortran', x)
                for x in direct_files]
ext1 = Extension(name='dragonfly.utils.direct_fortran.direct',
                 sources=direct_paths)

setup_options = dict(
    name='dragonfly-opt',
    version="0.1.1",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url='https://github.com/dragonfly/dragonfly/',
    license='MIT',
    author_email='kandasamy@cs.cmu.edu',
    packages=find_packages(exclude=('examples*', 'dragonfly.test_data*')),
    scripts=['bin/dragonfly-script.py'],
    install_requires=[
      'future',
      'numpy',
      'scipy',
      'six',
    ],
    extras_require={
        'extra': ['cython', 'POT'],
    },
    classifiers=[
      "Development Status :: 3 - Alpha",
      "Intended Audience :: Developers",
      "Intended Audience :: Education",
      "Intended Audience :: Science/Research",
      "License :: OSI Approved :: MIT License",
      "Operating System :: MacOS",
      "Operating System :: Microsoft :: Windows",
      "Operating System :: POSIX :: Linux",
      "Operating System :: Unix",
      "Programming Language :: Python :: 2",
      "Programming Language :: Python :: 3",
      "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)

try:
    # Try building the Fortran extension.
    setup(
        ext_modules=[ext1],
        cmdclass={"build_ext": construct_build_ext(old_build_ext)},
        **setup_options
    )
except BuildFailed:
    print("")
    print("*" * 80)
    print("Fortran compilation failed. Falling back on pure Python version.")
    print("*" * 80)
    setup(**setup_options)




"""
  Harness for black box optimisation.
  -- kandasamy@cs.cmu.edu
"""
from __future__ import division

# pylint: disable=abstract-class-little-used
# pylint: disable=invalid-name


import numpy as np
# Local imports
from ..exd.exd_core import ExperimentDesigner, exd_core_args
from ..utils.option_handler import load_options

blackbox_opt_args = exd_core_args


class CalledMFOptimiserWithSFCaller(Exception):
  """ An exception to handle calling of a multi-fidelity optimiser with a single fidelity
      caller.
  """
  def __init__(self, optimiser, func_caller):
    """ Constructor. """
    err_msg = ('Called optimiser %s with func_caller %s. func_caller needs to be ' +
               'multi-fidelity.')%(str(optimiser), str(func_caller))
    super(CalledMFOptimiserWithSFCaller, self).__init__(err_msg)


class BlackboxOptimiser(ExperimentDesigner):
  """ Blackbox Optimiser Class. """
  # pylint: disable=attribute-defined-outside-init

  def __init__(self, func_caller, worker_manager, model=None, options=None,
               reporter=None):
    """ Constructor. """
    self.func_caller = func_caller
    self.domain = self.func_caller.domain
    super(BlackboxOptimiser, self).__init__(func_caller, worker_manager, model,
                                            options, reporter)

  def _exd_child_set_up(self):
    """ Set up for the optimisation. """
    if self.func_caller.is_mf():
      self.num_fidel_to_opt_calls = 0
    self._blackbox_optimise_set_up()
    self._opt_method_set_up()
    self.prev_eval_vals = [] # for optimiser, prev_eval_vals

  def _blackbox_optimise_set_up(self):
    """ Set up for black-box optimisation. """
    # Initialise optimal value and point
    self.curr_opt_val = -np.inf
    self.curr_opt_point = None
    self.curr_true_opt_val = -np.inf
    self.curr_true_opt_point = None
    # Set up history
    self.history.query_vals = []
    self.history.query_true_vals = []
    self.history.curr_opt_vals = []
    self.history.curr_opt_points = []
    self.history.curr_true_opt_vals = []
    self.history.curr_true_opt_points = []
    if self.func_caller.is_mf():
      self.history.query_at_fidel_to_opts = []
    # Set up attributes to be copied from history
    self.to_copy_from_qinfo_to_history['val'] = 'query_vals'
    self.to_copy_from_qinfo_to_history['true_val'] = 'query_true_vals'
    # Set up previous evaluations
    self.history.prev_eval_points = []
    self.history.prev_eval_vals = []
    self.prev_eval_vals = []
    self.prev_eval_true_vals = []

  def _opt_method_set_up(self):
    """ Any set up for the specific optimisation method. """
    raise NotImplementedError('Implement in Optimisation Method class.')

  def _get_problem_str(self):
    """ Description of the problem. """
    return 'Optimisation'

  # Book-keeping ----------------------------------------------------------------
  def _exd_child_update_history(self, qinfo):
    """ Updates to the history specific to optimisation. """
    # Update the best point/val
    # check fidelity
    if self.func_caller.is_mf():
      eval_fidel = qinfo.fidel if hasattr(qinfo, 'fidel') else \
                     self.func_caller.fidel_to_opt
      query_is_at_fidel_to_opt = self.func_caller.is_fidel_to_opt(eval_fidel)
      self.history.query_at_fidel_to_opts.append(query_is_at_fidel_to_opt)
      self.num_fidel_to_opt_calls += query_is_at_fidel_to_opt
      self._update_opt_point_and_val(qinfo, query_is_at_fidel_to_opt)
    else:
      self._update_opt_point_and_val(qinfo)
    # Now add to history
    self.history.curr_opt_vals.append(self.curr_opt_val)
    self.history.curr_opt_points.append(self.curr_opt_point)
    self.history.curr_true_opt_vals.append(self.curr_true_opt_val)
    self.history.curr_true_opt_points.append(self.curr_true_opt_point)
    # Any method specific updating
    self._opt_method_update_history(qinfo)

  def _update_opt_point_and_val(self, qinfo, query_is_at_fidel_to_opt=None):
    """ Updates the optimum point and value according the data in qinfo.
        For single fidelity methods we update if qinfo.val is larger than the current
        value. For multi-fidelity methods, we do the same but also check if qinfo.fidel
        is the same as opt_fidel. Can be overridden by a child class if you want to do
        anything differently.
    """
    if query_is_at_fidel_to_opt is not None:
      if not query_is_at_fidel_to_opt:
        # if the fidelity queried at is not fidel_to_opt, then return
        return
    # Optimise curr_opt_val and curr_true_opt_val
    if qinfo.val > self.curr_opt_val:
      self.curr_opt_val = qinfo.val
      self.curr_opt_point = qinfo.point
    if qinfo.true_val > self.curr_true_opt_val:
      self.curr_true_opt_val = qinfo.true_val
      self.curr_true_opt_point = qinfo.point

  def _opt_method_update_history(self, qinfo):
    """ Any updates to the history specific to the method. """
    pass # Pass by default. Not necessary to override.

  def _get_exd_child_header_str(self):
    """ Header for black box optimisation. """
    ret = 'curr_max=<current_maximum_value>'
    if self.func_caller.is_mf():
      ret += ', f2o=<#queries_at_highest_fidelity>' + \
             '(<#queries_at_highest_fidelity_in_last_20_iterations>)'
    ret += self._get_opt_method_header_str()
    return ret

  @classmethod
  def _get_opt_method_header_str(cls):
    """ Header for optimisation method. """
    return ''

  def _get_exd_child_report_results_str(self):
    """ Returns a string describing the progress in optimisation. """
    best_val_str = 'curr_max=%0.5f'%(self.curr_opt_val)
    if self.func_caller.is_mf():
      window_length = 20
      window_queries_at_f2o = self.history.query_at_fidel_to_opts[-window_length:]
      fidel_to_opt_str = ', #f2o=%d(%d/%d)'%(self.num_fidel_to_opt_calls,
        sum(window_queries_at_f2o), window_length)
    else:
      fidel_to_opt_str = ''
    opt_method_str = self._get_opt_method_report_results_str()
    return best_val_str + fidel_to_opt_str + opt_method_str + ', '

  def _get_opt_method_report_results_str(self):
    """ Any details to include in a child method when reporting results.
        Can be overridden by a child class.
    """
    #pylint: disable=no-self-use
    return ''

  def _exd_child_handle_prev_evals(self):
    """ Handles pre-evaluations. """
    for qinfo in self.options.prev_evaluations.qinfos:
      if self.func_caller.is_mf():
        eval_fidel = qinfo.fidel if hasattr(qinfo, 'fidel') else \
                     self.func_caller.fidel_to_opt
        self.prev_eval_fidels.append(eval_fidel)
        query_is_at_fidel_to_opt = self.func_caller.is_fidel_to_opt(eval_fidel)
        self._update_opt_point_and_val(qinfo, query_is_at_fidel_to_opt)
      else:
        self._update_opt_point_and_val(qinfo)
      self.prev_eval_points.append(qinfo.point)
      self.prev_eval_vals.append(qinfo.val)
    self.history.prev_eval_points = self.prev_eval_points
    self.history.prev_eval_vals = self.prev_eval_vals

  def _child_run_experiments_initialise(self):
    """ Handles any initialisation before running experiments. """
    self._opt_method_optimise_initalise()

  def _opt_method_optimise_initalise(self):
    """ Any routine to run for a method just before optimisation routine. """
    pass # Pass by default. Not necessary to override.

  def optimise(self, max_capital):
    """ Calling optimise with optimise the function. A wrapper for run_experiments from
        BlackboxExperimenter. """
    return self.run_experiments(max_capital)

  def _get_final_return_quantities(self):
    """ Return the curr_opt_val, curr_opt_point and history. """
    return self.curr_opt_val, self.curr_opt_point, self.history


# An initialiser class for Optimisers ----------------------------------------------------
# Can be used to evaluate just a set of initial points.
class OptInitialiser(BlackboxOptimiser):
  """ An initialiser class. """
  # pylint: disable=no-self-use

  def __init__(self, func_caller, worker_manager, get_initial_qinfos=None,
               initialisation_capital=None, options=None, reporter=None):
    """ Constructor. """
    if options is None:
      options = load_options(blackbox_opt_args, reporter=reporter)
    super(OptInitialiser, self).__init__(func_caller, worker_manager, model=None,
                                         options=options, reporter=reporter)
    self.options.max_num_steps = 0
    self.options.get_initial_qinfos = get_initial_qinfos
    self.options.init_capital = initialisation_capital

  def _opt_method_set_up(self):
    """ Any set up for the specific optimisation method. """
    pass

  def _get_method_str(self):
    """ Return a string describing the method. """
    return 'initialiser'

  def is_asynchronous(self):
    """ Returns true if asynchronous."""
    return True

  def is_an_mf_method(self):
    """ Returns True if the method is a multi-fidelity method. """
    return self.func_caller.is_mf()

  def _get_exd_child_report_results_str(self):
    """ Returns a string for the specific child method describing the progress. """
    return ''

  def _exd_child_handle_prev_evals(self):
    """ Handles pre-evaluations. """
    raise ValueError('No reason to call this method in an initialiser.')

  def _get_initial_qinfos(self, num_init_evals):
    """ Returns the initial qinfos. Can be overridden by a child class. """
    # pylint: disable=unused-argument
    # pylint: disable=no-self-use
    raise ValueError('No reason to call this method in an initialiser.')

  def _child_run_experiments_initialise(self):
    """ Handles any initialisation before running experiments. """
    pass

  def _determine_next_query(self):
    """ Determine the next point for evaluation. """
    raise ValueError('No reason to call this method in an initialiser.')

  def _determine_next_batch_of_queries(self, _):
    """ Determine the next batch of eavluation points. """
    raise ValueError('No reason to call this method in an initialiser.')

  def _add_data_to_model(self, qinfos):
    """ Adds data to model. """
    pass

  def _child_build_new_model(self):
    """ Builds a new model. """
    pass

  def initialise(self):
    """ Initialise. """
    return self.optimise(0)




"""
  Implements various kernels.
  -- kandasamy@cs.cmu.edu
  -- kvysyara@andrew.cmu.edu
"""

from __future__ import division

# pylint: disable=invalid-name
# pylint: disable=abstract-method

import numpy as np
# Local imports
from ..utils.ancillary_utils import get_list_of_floats_as_str
from ..utils.general_utils import dist_squared, get_idxs_from_list_of_lists, \
                                pairwise_hamming_kernel


def get_list_of_hyperparams_for_kernel(kernel_type):
  """ Returns the list of hyper-parameters for each kernel.
      It returns a 3-tuple of lists of equal length.
      The first is a list of strings describing each hyper-parameter.
      The second is a list of integers indicating the size/dimensionality of each
      hyper-parameter. If any element in this list is 'dim', it means the dimensionality
      is the dimensionality of the input space.
      The third is a list of strings indicating if they are float or integral. If its a
      list instead of a string, then it means the hyper-parameter is discrete with the
      elements in the list denoting possible values.
  """
  if kernel_type.lower() == 'se':
    return ['scale', 'dim_bandwidths'], [1, 'dim'], ['float', 'float']
  elif kernel_type.lower() == 'matern':
    return (['scale', 'dim_bandwidths', 'nu'], [1, 'dim', 1],
            ['float', 'float', [0.5, 1.5, 2.5]])
  elif kernel_type.lower() == 'espse':
    return ['esp_order', 'scale', 'dim_bandwidths'], [1, 'dim', 1], \
           ['float', 'float', 'int']
  elif kernel_type.lower() == 'espmatern':
    return (['esp_order', 'scale', 'dim_bandwidths', 'nu'], [1, 'dim', 1, 1],
            ['float', 'float', [0.5, 1.5, 2.5], 'int'])
  elif kernel_type.lower() == 'expdecay':
    raise NotImplementedError('Not implemented this function for ExpDecayKernel yet!')
  else:
    raise ValueError('Unidentified kernel type %s.'%(kernel_type))


# Utilities -------
def _get_se_matern_scale_bw_strs(kern):
  """ Gets descriptions of the bandwidths and scales for the SE and matern kernels. """
  if kern.dim > 6:
    bw_str = 'avg-bw: %0.4f'%(kern.hyperparams['dim_bandwidths'].mean())
  else:
    bw_str = 'bws:[' + ' '.join(['%0.2f'%(dbw) for dbw in
                                 kern.hyperparams['dim_bandwidths']]) + ']'
  scale_str = 'sc:%0.4f'%(kern.hyperparams['scale'])
  return scale_str, bw_str


class Kernel(object):
  """ A kernel class. """

  def __init__(self):
    """ Constructor. """
    super(Kernel, self).__init__()
    self.hyperparams = {}

  def is_guaranteed_psd(self):
    """ The child class should implement this method to indicate whether it is
        guaranteed to be PSD. """
    raise NotImplementedError('Implement in a child class.')

  def __call__(self, X1, X2=None):
    """ Evaluates the kernel by calling evaluate. """
    return self.evaluate(X1, X2)

  def evaluate(self, X1, X2=None):
    """ Evaluates kernel values between X1 and X2 and returns an n1xn2 kernel matrix.
        This is a wrapper for _child_evaluate.
    """
    X2 = X1 if X2 is None else X2
    if len(X1) == 0 or len(X2) == 0:
      return np.zeros((len(X1), len(X2)))
    return self._child_evaluate(X1, X2)

  def _child_evaluate(self, X1, X2):
    """ Evaluates kernel values between X1 and X2 and returns an n1xn2 kernel matrix.
        This is to be implemented in a child kernel.
    """
    raise NotImplementedError('Implement in a child class.')

  def evaluate_from_dists(self, dists):
    """ Evaluates the kernel from pairwise distances. """
    raise NotImplementedError('Implement in a child class.')

  def set_hyperparams(self, **kwargs):
    """ Set hyperparameters here. """
    self.hyperparams = kwargs

  def add_hyperparams(self, **kwargs):
    """ Set additional hyperparameters here. """
    for key, value in kwargs.items():
      self.hyperparams[key] = value

  def get_effective_norm(self, X, order=None, *args, **kwargs):
    """ Gets the effective norm scaled by bandwidths. """
    raise NotImplementedError('Implement in a child class.')

  def compute_std_slack(self, X1, X2):
    """ Computes a bound on the maximum standard deviation between X1 and X2. """
    raise NotImplementedError('Implement in a child class.')

  def change_smoothness(self, factor):
    """ Decreases smoothness by the factor given. """
    raise NotImplementedError('Implement in a child class.')

  def gradient(self, param, X1, X2=None, *args):
    """ Computes gradient of kernel w.r.t the param """
    if len(X1) == 0 or len(X2) == 0:
      return np.zeros((len(X1), len(X2)))
    X2 = X1 if X2 is None else X2
    return self._child_gradient(param, X1, X2, *args)

  def _child_gradient(self, param, X1, X2, param_num=None):
    """ Computes gradient of kernel w.r.t the param """
    raise NotImplementedError('Implement in a child class.')

  def __str__(self):
    """ Returns a string representation of the kernel. """
    return '%s:: %s'%(type(self), str(self.hyperparams))


class SEKernel(Kernel):
  """ Squared exponential kernel. """

  def __init__(self, dim, scale=None, dim_bandwidths=None):
    """ Constructor. dim is the dimension. """
    super(SEKernel, self).__init__()
    self.dim = dim
    self.set_se_hyperparams(scale, dim_bandwidths)

  def is_guaranteed_psd(self):
    """ The child class should implement this method to indicate whether it is
        guaranteed to be PSD. """
    return True

  def set_dim_bandwidths(self, dim_bandwidths):
    """ Sets the bandwidth for each dimension. """
    if dim_bandwidths is not None:
      if len(dim_bandwidths) != self.dim:
        raise ValueError('Dimension of dim_bandwidths should be the same as dimension.')
      dim_bandwidths = np.array(dim_bandwidths).T
    self.add_hyperparams(dim_bandwidths=dim_bandwidths)

  def set_single_bandwidth(self, bandwidth):
    """ Sets the bandwidth of all dimensions to be the same value. """
    dim_bandwidths = None if bandwidth is None else [bandwidth] * self.dim
    self.set_dim_bandwidths(dim_bandwidths)

  def set_scale(self, scale):
    """ Sets the scale parameter for the kernel. """
    self.add_hyperparams(scale=scale)

  def set_se_hyperparams(self, scale, dim_bandwidths):
    """ Sets both the scale and the dimension bandwidths for the SE kernel. """
    self.set_scale(scale)
    if hasattr(dim_bandwidths, '__len__'):
      self.set_dim_bandwidths(dim_bandwidths)
    else:
      self.set_single_bandwidth(dim_bandwidths)

  def _child_evaluate(self, X1, X2):
    """ Evaluates the SE kernel between X1 and X2 and returns the gram matrix. """
    scaled_X1 = self.get_scaled_repr(X1)
    scaled_X2 = self.get_scaled_repr(X2)
    dist_sq = dist_squared(scaled_X1, scaled_X2)
    K = self.hyperparams['scale'] * np.exp(-dist_sq/2)
    return K

  def get_scaled_repr(self, X):
    """ Returns the scaled version of an input by the bandwidths. """
    return X/self.hyperparams['dim_bandwidths']

  def get_effective_norm(self, X, order=None, is_single=True):
    """ Gets the effective norm. That is the norm of X scaled by bandwidths. """
    # pylint: disable=arguments-differ
    scaled_X = self.get_scaled_repr(X)
    if is_single:
      return np.linalg.norm(scaled_X, ord=order)
    else:
      return np.array([np.linalg.norm(sx, ord=order) for sx in scaled_X])

  def compute_std_slack(self, X1, X2):
    """ Computes a bound on the maximum standard deviation diff between X1 and X2. """
    k_12 = np.array([float(self.evaluate(X1[i].reshape(1, -1), X2[i].reshape(1, -1)))
                     for i in range(len(X1))])
    return np.sqrt(self.hyperparams['scale'] - k_12)

  def change_smoothness(self, factor):
    """ Decreases smoothness by the given factor. """
    self.hyperparams['dim_bandwidths'] *= factor

  def _child_gradient(self, param, X1, X2, param_num=None):
    """ Computes gradient of kernel w.r.t the param """
    scaled_X1 = self.get_scaled_repr(X1)
    scaled_X2 = self.get_scaled_repr(X2)
    dist_sq = dist_squared(scaled_X1, scaled_X2)
    if param == 'scale':
      return self.hyperparams['scale'] * np.exp(-dist_sq/2)
    elif param == 'same_dim_bandwidths':
      dist_sq_dv = dist_sq/(self.hyperparams['dim_bandwidths'][0, 0])
      return self.hyperparams['scale'] * (np.multiply(dist_sq_dv, np.exp(-dist_sq/2)))
    else:
      dim_X1 = np.expand_dims(scaled_X1[:, param_num], axis=1)
      dim_X2 = np.expand_dims(scaled_X2[:, param_num], axis=1)
      dim_sq = dist_squared(dim_X1, dim_X2)
      dim_sq = dim_sq/(self.hyperparams['dim_bandwidths'][0, param_num])
      return self.hyperparams['scale'] * (np.multiply(dim_sq, np.exp(-dist_sq/2)))

  def __str__(self):
    """ Returns a string representation of the kernel. """
    scale_str, bw_str = _get_se_matern_scale_bw_strs(self)
    return 'SE: ' + scale_str + ' ' + bw_str


class MaternKernel(Kernel):
  """ The Matern class of kernels. """
  # pylint: disable=abstract-method

  def __init__(self, dim, nu=None, scale=None, dim_bandwidths=None):
    """ Constructor. dim is the dimension. """
    super(MaternKernel, self).__init__()
    self.dim = dim
    self.p = None
    self.norm_constant = None
    self.set_matern_hyperparams(nu, scale, dim_bandwidths)

  def is_guaranteed_psd(self):
    """ The child class should implement this method to indicate whether it is
        guaranteed to be PSD. """
    return True

  def set_matern_hyperparams(self, nu, scale, dim_bandwidths):
    """ Sets the parameters of the matern kernel. """
    if nu%1 != 0.5:
      raise ValueError('Matern kernel: nu has to be p + 0.5 where p is an integer.')
    self.add_hyperparams(nu=nu)
    self.add_hyperparams(scale=scale)
    dim_bandwidths = dim_bandwidths if hasattr(dim_bandwidths, '__len__') else \
                     [dim_bandwidths] * self.dim
    dim_bandwidths = np.array(dim_bandwidths).T
    self.add_hyperparams(dim_bandwidths=dim_bandwidths)
    self.p = int(nu)
    self.norm_constant = 1.0 / self._eval_kernel_values_unnormalised(0)

  def get_scaled_repr(self, X):
    """ Returns scaled versions of the input by the bandwidths. """
    return X/self.hyperparams['dim_bandwidths']

  def _eval_kernel_values_unnormalised(self, dist):
    """ Given the distances, evaluates the unnormalised kernel value. """
    unnorm_K = np.zeros(dist.shape) if isinstance(dist, np.ndarray) else 0
    for i in range(self.p+1):
      coeff = np.math.factorial(self.p+i)/(np.math.factorial(i) *
                                           np.math.factorial(self.p-i))
      mult_mat = (np.sqrt(8 * self.hyperparams['nu']) * dist)
      unnorm_K += coeff * (mult_mat)**(self.p - i)
    # Finally multiply by the leading exponential
    unnorm_K *= (np.math.gamma(self.p+1) / np.math.gamma(2*self.p+1) *
                 np.exp(-np.sqrt(2 * self.hyperparams['nu']) * dist))
    return unnorm_K

  def _eval_grad_kernel_values_unnormalised(self, dist, dist_dv):
    """ Given the distances, evaluates the gradient of unnormalised kernel value. """
    unnorm_K = np.zeros(dist.shape) if isinstance(dist, np.ndarray) else 0
    unnorm_K_dv = np.zeros(dist_dv.shape) if isinstance(dist_dv, np.ndarray) else 0
    for i in range(self.p+1):
      coeff = np.math.factorial(self.p+i)/(np.math.factorial(i) *
                                           np.math.factorial(self.p-i))
      mult_mat = (np.sqrt(8 * self.hyperparams['nu']) * dist)
      unnorm_K += coeff * (mult_mat)**(self.p - i)
      if self.p - i > 0:
        unnorm_K_dv += (np.sqrt(8 * self.hyperparams['nu'])) * (self.p - i) * \
                       coeff * ((mult_mat)**(self.p - i - 1)) * dist_dv
    # Finally multiply by the leading exponential
    unnorm_K_dv *= (np.math.gamma(self.p+1) / np.math.gamma(2*self.p+1) *
                    np.exp(-np.sqrt(2 * self.hyperparams['nu']) * dist))
    unnorm_K *= (np.math.gamma(self.p+1) / np.math.gamma(2*self.p+1) *
                 np.exp(-np.sqrt(2 * self.hyperparams['nu']) * dist) *
                 (-np.sqrt(2 * self.hyperparams['nu']) * dist_dv))
    return unnorm_K + unnorm_K_dv

  def _child_evaluate(self, X1, X2):
    """ Evaluates the SE kernel between X1 and X2 and returns the gram matrix. """
    scaled_X1 = self.get_scaled_repr(X1)
    scaled_X2 = self.get_scaled_repr(X2)
    dist = np.sqrt(dist_squared(scaled_X1, scaled_X2))
    unnorm_kernel_vals = self._eval_kernel_values_unnormalised(dist)
    K = self.hyperparams['scale'] * self.norm_constant * unnorm_kernel_vals
    return K

  def _child_gradient(self, param, X1, X2, param_num=None):
    """ Computes gradient of kernel w.r.t the param. """
    scaled_X1 = self.get_scaled_repr(X1)
    scaled_X2 = self.get_scaled_repr(X2)
    dist = np.sqrt(dist_squared(scaled_X1, scaled_X2))
    if param == 'scale':
      unnorm_kernel_vals = self._eval_kernel_values_unnormalised(dist)
      return self.hyperparams['scale'] * self.norm_constant * unnorm_kernel_vals
    elif param == 'same_dim_bandwidths':
      dist_dv = -(dist/(self.hyperparams['dim_bandwidths'][0, 0]))
      unnorm_kernel_vals = self._eval_grad_kernel_values_unnormalised(dist, dist_dv)
      return self.hyperparams['scale'] * self.norm_constant * unnorm_kernel_vals
    else:
      dim_X1 = np.expand_dims(scaled_X1[:, param_num], axis=1)
      dim_X2 = np.expand_dims(scaled_X2[:, param_num], axis=1)
      np.fill_diagonal(dist, 1.0)
      dist_dv = 1/dist
      np.fill_diagonal(dist, 0.0)
      dim_sq = dist_squared(dim_X1, dim_X2)
      dist_dv *= -(dim_sq/(self.hyperparams['dim_bandwidths'][0, param_num]))
      unnorm_kernel_vals = self._eval_grad_kernel_values_unnormalised(dist, dist_dv)
      return self.hyperparams['scale'] * self.norm_constant * unnorm_kernel_vals

  def __str__(self):
    """ Returns a string representation of the kernel. """
    scale_str, bw_str = _get_se_matern_scale_bw_strs(self)
    nu_str = 'nu=%0.1f'%(self.hyperparams['nu'])
    return 'Matern: ' + nu_str + ' ' + scale_str + ' ' + bw_str


class PolyKernel(Kernel):
  """ The polynomial kernel. """
  # pylint: disable=abstract-method

  def __init__(self, dim, order, scale, dim_scalings=None):
    """ Constructor. """
    super(PolyKernel, self).__init__()
    self.dim = dim
    self.set_poly_hyperparams(order, scale, dim_scalings)

  def is_guaranteed_psd(self):
    """ The child class should implement this method to indicate whether it is
        guaranteed to be PSD. """
    return True

  def set_order(self, order):
    """ Sets the order of the polynomial. """
    self.add_hyperparams(order=order)

  def set_scale(self, scale):
    """ Sets the scale of the kernel. """
    self.add_hyperparams(scale=scale)

  def set_dim_scalings(self, dim_scalings):
    """ Sets the scaling for each dimension in the polynomial kernel. This will be a
        dim+1 dimensional vector.
    """
    if dim_scalings is not None:
      if len(dim_scalings) != self.dim:
        raise ValueError('Dimension of dim_scalings should be dim.')
      dim_scalings = np.array(dim_scalings)
    self.add_hyperparams(dim_scalings=dim_scalings)

  def set_single_scaling(self, scaling):
    """ Sets the same scaling for all dimensions. """
    if scaling is None:
      self.set_dim_scalings(None)
    else:
      self.set_dim_scalings([scaling] * self.dim)

  def set_poly_hyperparams(self, order, scale, dim_scalings):
    """Sets the hyper parameters. """
    self.set_order(order)
    self.set_scale(scale)
    if hasattr(dim_scalings, '__len__'):
      self.set_dim_scalings(dim_scalings)
    else:
      self.set_single_scaling(dim_scalings)

  def _child_evaluate(self, X1, X2):
    """ Evaluates the polynomial kernel and returns and the gram matrix. """
    X1 = X1 * self.hyperparams['dim_scalings']
    X2 = X2 * self.hyperparams['dim_scalings']
    K = self.hyperparams['scale'] * ((X1.dot(X2.T) + 1)**self.hyperparams['order'])
    return K

  def __str__(self):
    """ Returns a string representation of the kernel. """
    dim_scalings_str = ','.join(['%0.2f'%(elem) for elem in
                                 self.hyperparams['dim_scalings']])
    return 'Poly: d=%d, scale=%0.2f, %s'%(self.hyperparams['order'], \
            self.hyperparams['scale'], dim_scalings_str)


class ExpDecayKernel(Kernel):
  """ A kernel for exponentially decaying functions. Taken from
      Freeze-Thaw Bayesian Optimization, Swersky et al.
  """
  # TODO: write unit tests for ExpDecay Kernel

  def __init__(self, dim, scale=None, offset=None, powers=None):
    """ Constructor. """
    super(ExpDecayKernel, self).__init__()
    self.dim = dim
    if not hasattr(powers, '__iter__'):
      powers = [powers] * dim
    self.set_hyperparams(scale=scale, offset=offset, powers=powers)

  def is_guaranteed_psd(self):
    """ The child class should implement this method to indicate whether it is
        guaranteed to be PSD. """
    return True

  def _child_evaluate(self, X1, X2):
    """ Evaluates the child kernel between X1 and X2 and returns the gram matrix. """
    if isinstance(X1, list):
      X1 = np.array(X1)
    if isinstance(X2, list):
      X2 = np.array(X2)
    powers = self.hyperparams['powers']
    ret = self.hyperparams['scale'] * np.ones((X1.shape[0], X2.shape[0]))
    for dim_idx in range(self.dim):
      dim_X1 = X1[:, dim_idx]
      dim_X2 = X2[:, dim_idx]
      dim_kernel = 1 / (1 + np.add.outer(dim_X1, dim_X2)) ** powers[dim_idx]
      ret *= dim_kernel
    ret += self.hyperparams['offset']
    return ret

  def __str__(self):
    """ Return a string representation. """
    return 'ExpDec: sc=%0.3f, offset=%0.3f, pow=%s'%(self.hyperparams['scale'], \
        self.hyperparams['offset'], get_list_of_floats_as_str(self.hyperparams['powers']))


class HammingKernel(Kernel):
  """ The Hamming kernel. """

  def __init__(self, dim_weights):
    """ Constructor. """
    super(HammingKernel, self).__init__()
    if isinstance(dim_weights, (int, float)):
      dim_weights = np.ones((dim_weights,))/float(dim_weights)
    dim_weights = np.array(dim_weights)
    self.set_hyperparams(dim_weights=dim_weights)

  def is_guaranteed_psd(self):
    """ Returns true. """
    return True

  def _child_evaluate(self, X1, X2):
    """ Evaluates the child kernel. """
    return pairwise_hamming_kernel(X1, X2, self.hyperparams['dim_weights'])

  def __str__(self):
    """ Return a string representation. """
    return 'Hamming: wts=%s'%(get_list_of_floats_as_str(self.hyperparams['dim_weights']))


# Derivative Kernels ====================================================================
class AdditiveKernel(Kernel):
  """Implements an Additive kernel on Euclidean Spaces with non-overlapping groups. """
  # pylint: disable=abstract-method

  def __init__(self, scale, kernel_list, groupings):
    """ Constructor:
        kernel_list is a list of kernel objects.
        groupings is a list of list first dimension is the group second dimension is the
        set of dimensions in the group.
    """
    if len(kernel_list) != len(groupings):
      raise ValueError("number of kernels do not correspond to number of groups.")
    super(AdditiveKernel, self).__init__()
    self.kernel_list = kernel_list
    self.groupings = groupings
    self.add_hyperparams(scale=scale)
    self.dim = sum([kern.dim for kern in self.kernel_list])

  def is_guaranteed_psd(self):
    """ The child class should implement this method to indicate whether it is
        guaranteed to be PSD. Returns True if each kernel is is_guaranteed_psd()"""
    return all([kern.is_guaranteed_psd() for kern in self.kernel_list])

  def _child_evaluate(self, X1, X2):
    "computes the gram matrix"
    # Convert them to arrays so that we can index the groups out.
    X1 = np.array(X1)
    X2 = np.array(X2)
    n1 = X1.shape[0]
    n2 = X2.shape[0]
    result = np.zeros((n1, n2))
    for kernel, group in zip(self.kernel_list, self.groupings):
      result += kernel(X1[:, group], X2[:, group])
    return self.hyperparams['scale'] * result

  def __str__(self):
    """ Returns a string representation of the kernel. """
    kernels_str_list = ['%s(%s)'%(grp, kern) for (grp, kern) in
                        zip(self.groupings, self.kernel_list)]
    kernels_str = ', '.join(kernels_str_list)
    return 'ADD scale=%0.2f, '%(self.hyperparams['scale']) + kernels_str


class CartesianProductKernel(Kernel):
  """ Implements a product kernel which takes the cartesian product of all kernels in
      a kernel list.
  """

  def __init__(self, scale, kernel_list):
    """ kernel_list is a list of kernel objects.
        We will assume that when computing the kernel, each X is a list of secondary
        lists where the j'th in a secondary list corresponds to the j'th kernel.
    """
    super(CartesianProductKernel, self).__init__()
    self.kernel_list = kernel_list
    self.num_kernels = len(kernel_list)
    self.add_hyperparams(scale=scale)

  def is_guaranteed_psd(self):
    """ The child class should implement this method to indicate whether it is
        guaranteed to be PSD. Returns True if each kernel is is_guaranteed_psd()"""
    return all([kern.is_guaranteed_psd() for kern in self.kernel_list])

  def _child_evaluate(self, X1, X2):
    """ Computes the Gram Matrix. """
    n1 = len(X1)
    n2 = len(X2)
    result = self.hyperparams['scale'] * np.ones((n1, n2))
    for idx, kern in enumerate(self.kernel_list):
      curr_X1 = get_idxs_from_list_of_lists(X1, idx)
      curr_X2 = get_idxs_from_list_of_lists(X2, idx)
      result *= kern(curr_X1, curr_X2)
    return result

  def __str__(self):
    """ Returns a string representation of the kernel. """
    kernels_str = ', '.join([str(kern) for kern in self.kernel_list])
    return 'DomProd scale=%0.2f, '%(self.hyperparams['scale']) + kernels_str


class CoordinateProductKernel(Kernel):
  """ Implements a coordinatewise product kernel. """
  # pylint: disable=abstract-method

  def __init__(self, dim, scale, kernel_list=None, coordinate_list=None):
    """ Constructor.
        kernel_list is a list of n Kernel objects. coordinate_list is a list of n lists
        each indicating the coordinates each kernel in kernel_list should be applied to.
    """
    super(CoordinateProductKernel, self).__init__()
    self.dim = dim
    self.add_hyperparams(scale=scale)
    self.kernel_list = kernel_list
    self.coordinate_list = coordinate_list

  def set_kernel_list(self, kernel_list):
    """ Sets a new list of kernels. """
    self.kernel_list = kernel_list

  def is_guaranteed_psd(self):
    """ The child class should implement this method to indicate whether it is
        guaranteed to be PSD. Returns True if each kernel is is_guaranteed_psd()"""
    return all([kern.is_guaranteed_psd() for kern in self.kernel_list])

  def set_new_kernel(self, kernel_idx, new_kernel):
    """ Sets new_kernel to kernel_list[kernel_idx]. """
    self.kernel_list[kernel_idx] = new_kernel

  def set_kernel_hyperparams(self, kernel_idx, **kwargs):
    """ Sets the hyper-parameters for kernel_list[kernel_idx]. """
    self.kernel_list[kernel_idx].set_hyperparams(**kwargs)

  def _child_evaluate(self, X1, X2):
    """ Evaluates the combined kernel. """
    n1 = len(X1)
    n2 = len(X2)
    X1 = np.array(X1)
    X2 = np.array(X2)
    K = self.hyperparams['scale'] * np.ones((n1, n2))
    for idx, kernel in enumerate(self.kernel_list):
      X1_curr = X1[:, self.coordinate_list[idx]]
      X2_curr = X2[:, self.coordinate_list[idx]]
      K *= kernel(X1_curr, X2_curr)
    return K

  def __str__(self):
    """ Returns a string representation of the kernel. """
    kernels_str_list = ['%s(%s)'%(grp, kern) for (grp, kern) in
                        zip(self.coordinate_list, self.kernel_list)]
    kernels_str = ', '.join(kernels_str_list)
    return 'CoordProd scale=%0.2f, '%(self.hyperparams['scale']) + kernels_str


class ExpSumOfDistsKernel(Kernel):
  """ Given a function that returns a list of distances d1, d2, ... dk, this kernel
      takes the form exp(beta1*d1^p + beta2*d2^p + ... + betak*dk^p. """
  # pylint: disable=abstract-method

  def __init__(self, dist_computer, betas, scale, powers=1, num_dists=None,
               dist_is_hilbertian=False):
    """ Constructor.
          trans_dist_computer: Given two lists of networks X1 and X2, trans_dist_computer
            is a function which returns a list of n1xn2 matrices where ni=len(Xi).
    """
    super(ExpSumOfDistsKernel, self).__init__()
    self.num_dists = num_dists if num_dists is not None else len(betas)
    self.dist_computer = dist_computer
    betas = betas if hasattr(betas, '__iter__') else [betas] * self.num_dists
    powers = powers if hasattr(powers, '__iter__') else [powers] * self.num_dists
    self.add_hyperparams(betas=np.array(betas), powers=np.array(powers), scale=scale)
    self.dist_is_hilbertian = dist_is_hilbertian

  def is_guaranteed_psd(self):
    """ The child class should implement this method to indicate whether it is
        guaranteed to be PSD. """
    return self.dist_is_hilbertian

  def evaluate_from_dists(self, list_of_dists):
    """ Evaluates the kernel from pairwise distances. """
    sum_of_dists = _compute_raised_scaled_sum(list_of_dists, self.hyperparams['betas'], \
                                                            self.hyperparams['powers'])
    return self.hyperparams['scale'] * np.exp(-sum_of_dists)

  def _child_evaluate(self, X1, X2):
    """ Evaluates the kernel between X1 and X2. """
    list_of_dists = self.dist_computer(X1, X2)
    return self.evaluate_from_dists(list_of_dists)


class SumOfExpSumOfDistsKernel(Kernel):
  """ Given a function that returns a list of distances d1, d2, ... dk, this kernel
      takes the form alpha_1*exp(beta11*d1^p + ... + beta1k*d2^p) +
      alpha_2 * exp(beta_21 + ... + beta2k*dk^p) + ... .
  """

  def __init__(self, dist_computer, alphas, groups, betas, powers, num_dists=None,
               dist_is_hilbertian=False):
    super(SumOfExpSumOfDistsKernel, self).__init__()
    self.num_dists = num_dists if num_dists is not None else len(betas)
    self.dist_computer = dist_computer
    assert len(alphas) == len(groups)
    betas = betas if hasattr(betas, '__iter__') else [betas] * self.num_dists
    powers = powers if hasattr(powers, '__iter__') else [powers] * self.num_dists
    self.add_hyperparams(betas=np.array(betas), powers=np.array(powers), alphas=alphas,
                         groups=groups)
    self.dist_is_hilbertian = dist_is_hilbertian

  def is_guaranteed_psd(self):
    """ The child class should implement this method to indicate whether it is
        guaranteed to be PSD. """
    return self.dist_is_hilbertian

  def evaluate_from_dists(self, list_of_dists):
    """ Evaluates the kernel from pairwise distances. """
    individual_kernels = []
    for gi, group in enumerate(self.hyperparams['groups']):
      curr_list_of_dists = [list_of_dists[i] for i in group]
      curr_betas = [self.hyperparams['betas'][i] for i in group]
      curr_powers = [self.hyperparams['powers'][i] for i in group]
      curr_sum_of_dists = _compute_raised_scaled_sum(curr_list_of_dists,
                                                     curr_betas, curr_powers)
      individual_kernels.append(self.hyperparams['alphas'][gi] *
                                np.exp(-curr_sum_of_dists))
    return sum(individual_kernels)

  def _child_evaluate(self, X1, X2):
    list_of_dists = self.dist_computer(X1, X2)
    return self.evaluate_from_dists(list_of_dists)


class ESPKernel(Kernel):
  """ Implements the ESP kernel from Kandasamy, Yu 2016, 'Additive Approximations in High
      Dimensional Nonparametric Regression. """

  def __init__(self, scale, order, kernel_list):
    super(ESPKernel, self).__init__()
    self.dim = len(kernel_list)
    self.kernel_list = kernel_list
    self.add_hyperparams(scale=scale, order=order)

    if self.dim < 0:
      raise ValueError("dim cannot not be negative")
    if order > self.dim:
      raise ValueError("order must be less than or equal to dim")
    if order < 1:
      raise ValueError("order must be an integer between 1 and dim")

  def is_guaranteed_psd(self):
    """ The child class should implement this method to indicate whether it is
        guaranteed to be PSD. Returns True if each kernel is is_guaranteed_psd()"""
    return all([kern.is_guaranteed_psd() for kern in self.kernel_list])

  def _child_evaluate(self, X1, X2):
    "computes the gram matrix"
    # pylint: disable=too-many-locals
    order = self.hyperparams["order"]
    kernel_list = self.kernel_list
    scale = self.hyperparams["scale"]
    if isinstance(X1, list):
      X1 = np.array(X1)
    if isinstance(X2, list):
      X2 = np.array(X2)

    n1 = X1.shape[0]
    n2 = X2.shape[0]
    kernel_matrices = []
    for i in range(self.dim):
      kernel = kernel_list[i]
      kernel_matrices.append(kernel(X1[:, i:i+1], X2[:, i:i+1]))
    power_sum = [np.zeros((n1, n2)) for _ in range(order + 1)]

    ones = np.ones((n1, n2))
    power_sum[0] = ones
    for i in range(1, order + 1):
      for matrix in kernel_matrices:
        power_sum[i] += matrix ** i

    # elementary symmetric polynomials
    esp = [np.zeros((n1, n2)) for _ in range(order + 1)]
    esp[0] = ones
    for m in range(1, order + 1):
      for i in range(1, m + 1):
        esp[m] += ((-1) ** (i - 1)) * esp[m - i] * power_sum[i]
      esp[m] /= m

    return scale * esp[order]


class ESPKernelSE(ESPKernel):
  """Implements ESP kernel with SE kernel for each dimension"""
  def __init__(self, dim, scale, order, dim_bandwidths):
    kernel_list = []
    for i in range(dim):
      kernel_list.append(SEKernel(1, 1.0, np.asscalar(dim_bandwidths[i])))
    super(ESPKernelSE, self).__init__(scale, order, kernel_list)


class ESPKernelMatern(ESPKernel):
  """Implements ESP kernel with Matern kernel for each dimension"""
  def __init__(self, dim, nu, scale, order, dim_bandwidths):
    kernel_list = []
    for i in range(dim):
      kernel_list.append(MaternKernel(1, nu[i], 1.0, np.asscalar(dim_bandwidths[i])))
    super(ESPKernelMatern, self).__init__(scale, order, kernel_list)


# Ancillary functions for the ExpSumOfDists and SumOfExpSumOfDistsKernel classes ---------
def _compute_raised_scaled_sum(dist_arrays, betas, powers):
  """ Returns the distances raised to the powers and scaled by betas. """
  sum_of_dists = np.zeros(dist_arrays[0].shape)
  for idx, curr_dist in enumerate(dist_arrays):
    if powers[idx] == 1:
      raised_curr_dists = curr_dist
    else:
      raised_curr_dists = curr_dist ** powers[idx]
    sum_of_dists += betas[idx] * raised_curr_dists
  return sum_of_dists




"""
  Unittests for Cartesian product GP with NN domains. Testing these separately since
  we do not need to have nn dependencies in the gp module.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=no-member

import os
import numpy as np
import unittest
# Local
from ..gp.unittest_cartesian_product_gp import get_test_dataset, \
     gen_cpmfgp_test_data_from_config_file, \
     CPGPTestCaseDefinitions, CPMFGPTestCaseDefinitions
from ..nn import nn_examples
from ..test_data.syn_cnn_2.syn_cnn_2 import syn_cnn_2
from ..test_data.syn_cnn_2.syn_cnn_2_mf import syn_cnn_2_mf
from ..utils.base_test_class import BaseTestClass, execute_tests
from ..utils.general_utils import map_to_bounds


def get_cnns():
  """ Returns the initial pool for CNNs. """
  vgg_nets = [nn_examples.get_vgg_net(1),
              nn_examples.get_vgg_net(2),
              nn_examples.get_vgg_net(3),
              nn_examples.get_vgg_net(4)]
  blocked_cnns = [nn_examples.get_blocked_cnn(3, 1, 1), # 3
                  nn_examples.get_blocked_cnn(3, 2, 1), # 6
                  nn_examples.get_blocked_cnn(3, 3, 1), # 9
                  nn_examples.get_blocked_cnn(3, 4, 1), # 12
                  nn_examples.get_blocked_cnn(3, 5, 1), # 15
                  nn_examples.get_blocked_cnn(4, 4, 1), # 16
                 ]
  other_cnns = nn_examples.generate_cnn_architectures()
  ret = vgg_nets + blocked_cnns + other_cnns
  np.random.shuffle(ret)
  return ret


def get_initial_mlp_pool(class_or_reg):
  """ Returns the initial pool of MLPs. """
  blocked_mlps = [nn_examples.get_blocked_mlp(class_or_reg, 3, 2), # 6
                  nn_examples.get_blocked_mlp(class_or_reg, 4, 2), # 8
                  nn_examples.get_blocked_mlp(class_or_reg, 5, 2), # 10
                  nn_examples.get_blocked_mlp(class_or_reg, 3, 4), # 12
                  nn_examples.get_blocked_mlp(class_or_reg, 6, 2), # 12
                  nn_examples.get_blocked_mlp(class_or_reg, 8, 2), # 16
                  nn_examples.get_blocked_mlp(class_or_reg, 6, 3), # 18
                  nn_examples.get_blocked_mlp(class_or_reg, 10, 2), #20
                  nn_examples.get_blocked_mlp(class_or_reg, 4, 6), #24
                  nn_examples.get_blocked_mlp(class_or_reg, 8, 3), #24
                 ]
  other_mlps = nn_examples.generate_mlp_architectures()
  ret = blocked_mlps + other_mlps
  np.random.shuffle(ret)
  return ret


def gen_cpnnmfgp_test_data(num_tr_data, num_te_data):
  """ Generates data on all functions. """
  file_dir = os.path.dirname(os.path.realpath(__file__))
  test_data_dir = os.path.dirname(file_dir)
  test_problems = [
    (test_data_dir + '/test_data/syn_cnn_2/config_mf.json', syn_cnn_2_mf),
    ]
  ret = [gen_cpmfgp_test_data_from_config_file(cfn, rf, num_tr_data, num_te_data)
         for cfn, rf in test_problems]
  return ret


def get_cp_nn_gp_test_data():
  """ Create test data. """
  # pylint: disable=too-many-locals
  file_dir = os.path.dirname(os.path.realpath(__file__))
  test_data_dir = os.path.dirname(file_dir)
  ret = []
  n_train = 200
  n_test = 300
  # Dataset 1
  all_cnns = get_cnns()
  num_train = max(len(all_cnns), n_train)
  num_test = max(len(all_cnns), n_test)
  domain_file_name = test_data_dir + '/' + 'test_data/syn_cnn_2/config.json'
  func = syn_cnn_2
  x1_bounds = np.array([[0, 1], [0, 1], [10, 14]])
  x4_elems = [4, 10, 23, 45, 78, 87.1, 91.8, 99, 75.7, 28.1, 3.141593]
    # Create training set
  X_tr_0 = np.random.choice(all_cnns, num_train)
  X_tr_1 = map_to_bounds(np.random.random((num_train, len(x1_bounds))), x1_bounds)
  X_tr_2 = [[elem1, elem2] for (elem1, elem2) in zip(
             np.random.choice(['foo', 'bar'], num_train),
             np.random.choice(['foo', 'bar'], num_train))]
  X_tr_3 = np.random.choice(x4_elems, num_train)
  X_train = [[x0, x1, x2, [x3]] for (x0, x1, x2, x3) in \
             zip(X_tr_0, X_tr_1, X_tr_2, X_tr_3)]
    # Create test set
  X_te_0 = np.random.choice(all_cnns, num_test)
  X_te_1 = map_to_bounds(np.random.random((num_test, len(x1_bounds))), x1_bounds)
  X_te_2 = [[elem1, elem2] for (elem1, elem2) in zip(
             np.random.choice(['foo', 'bar'], num_test),
             np.random.choice(['foo', 'bar'], num_test))]
  X_te_3 = np.random.choice(x4_elems, num_test)
  X_test = [[x0, x1, x2, [x3]] for (x0, x1, x2, x3) in \
             zip(X_te_0, X_te_1, X_te_2, X_te_3)]
  ret.append(get_test_dataset(domain_file_name, func, X_train, X_test))
  return ret


class CPNNGPTestCase(CPGPTestCaseDefinitions, BaseTestClass):
  """ Unit tests for Cartesian product GPs with NN domains. """

  def setUp(self):
    """ Set up. """
    self.cpgp_datasets = get_cp_nn_gp_test_data()
    self.num_datasets = len(self.cpgp_datasets)


@unittest.skip
class CPMFGPTestCase(CPMFGPTestCaseDefinitions, BaseTestClass):
  """ Unit tests for Multi-fidelity Cartesian Product GPs. """

  def setUp(self):
    """ Set up. """
    num_tr_data = 50
    num_te_data = 50
    self.cpmfgp_datasets = gen_cpnnmfgp_test_data(num_tr_data, num_te_data)
    self.num_datasets = len(self.cpmfgp_datasets)



if __name__ == '__main__':
  execute_tests()




"""
  Monitors are used to monitor the progress of any algorithm. Here we implement
  the most basic monitor which will be inherited by monitors for specific processes.
  -- kandasamy@cs.cmu.edu
"""

import sys
import six
import io

def get_reporter(reporter):
  """ Returns a reporter based on what was passed as argument. If reporter is already
      a reporter then it is returned. Otherwise, an appropriate reporter is constructed
      and returned. """
  if six.PY2:
    file_types = file, io.IOBase
  else:
    file_types = (io.IOBase, )

  if isinstance(reporter, str):
    if reporter.lower() == 'default':
      reporter = BasicReporter()
    elif reporter.lower() == 'silent':
      reporter = SilentReporter()
    else:
      raise ValueError('If reporter is string, it should be "default" or "silent".')
  elif isinstance(reporter, file_types):
    return BasicReporter(reporter)
  elif reporter is None:
    reporter = SilentReporter()
  elif not isinstance(reporter, BasicReporter):
    raise ValueError('Pass either a string, BasicReporter or None for reporter. ')
  return reporter


class BasicReporter(object):
  """ The most basic monitor that implements printing etc. """

  def __init__(self, out=sys.stdout):
    """ Constructor. """
    self.out = out

  def write(self, msg, *_):
    """ Writes a message to stdout. """
    if self.out is not None:
      self.out.write(msg)
      self.out.flush()

  def writeln(self, msg, *args):
    """ Writes a message to stdout with a new line. """
    self.write(msg + '\n', *args)


class SilentReporter(BasicReporter):
  """ This reporter prints nothing. """
  def __init__(self):
    """ Constructor. """
    super(SilentReporter, self).__init__(None)




"""
  Defines a class for Multi-objective Blackbox Optimisation.
  -- bparia@cs.cmu.edu
  -- kandasamy@cs.cmu.edu
"""

# NB: In this file, the acronym MOO/moo refers to multi-objective optimisation. --KK

from __future__ import division

# pylint: disable=abstract-class-little-used
# pylint: disable=invalid-name


# Local imports
from ..exd.exd_core import ExperimentDesigner, exd_core_args
from ..exd.experiment_caller import MultiFunctionCaller, FunctionCaller
from ..utils.general_utils import update_pareto_set


multiobjective_opt_args = exd_core_args

_NO_MF_FOR_MOO_ERR_MSG = 'Multi-fidelity support has not been implemented yet' + \
                         ' for multi-objective optimisation.'


class MultiObjectiveOptimiser(ExperimentDesigner):
  """ Blackbox Optimiser Class. """
  # pylint: disable=attribute-defined-outside-init

  def __init__(self, multi_func_caller, worker_manager, model=None, options=None,
               reporter=None):
    """ Constructor. """
    assert isinstance(multi_func_caller, MultiFunctionCaller) and not \
           isinstance(multi_func_caller, FunctionCaller)
    self.multi_func_caller = multi_func_caller
    # If it is not a list, computing the non-dominated set is equivalent
    # to computing the maximum.
    self.domain = self.multi_func_caller.domain
    super(MultiObjectiveOptimiser, self).__init__(multi_func_caller, worker_manager,
                                                  model, options, reporter)

  def _exd_child_set_up(self):
    """ Set up for the optimisation. """
    if self.multi_func_caller.is_mf():
      # self.num_fidel_to_opt_calls = 0
      raise NotImplementedError(_NO_MF_FOR_MOO_ERR_MSG)
    self._moo_set_up()
    self._multi_opt_method_set_up()
    self.prev_eval_vals = [] # for optimiser, prev_eval_vals

  def _moo_set_up(self):
    """ Set up for black-box optimisation. """
    # Initialise optimal values and points
    # (Optimal point for MF problems is not defined)
    # (Instead a set of pareto optimal points will be maintained)
    self.curr_pareto_vals = []
    self.curr_pareto_points = []
    self.curr_true_pareto_vals = []
    self.curr_true_pareto_points = []
    # Set up history
    self.history.query_vals = []
    self.history.query_true_vals = []
    self.history.curr_pareto_vals = []
    self.history.curr_pareto_points = []
    self.history.curr_true_pareto_vals = []
    self.history.curr_true_pareto_points = []
    if self.multi_func_caller.is_mf():
      # self.history.query_at_fidel_to_opts = []
      raise NotImplementedError(_NO_MF_FOR_MOO_ERR_MSG)
    # Set up attributes to be copied from history
    self.to_copy_from_qinfo_to_history['val'] = 'query_vals'
    self.to_copy_from_qinfo_to_history['true_val'] = 'query_true_vals'
    # Set up previous evaluations
    self.history.prev_eval_points = []
    self.history.prev_eval_vals = []
    self.prev_eval_vals = []
    self.prev_eval_true_vals = []

  def _multi_opt_method_set_up(self):
    """ Any set up for the specific optimisation method. """
    raise NotImplementedError('Implement in Optimisation Method class.')

  def _get_problem_str(self):
    """ Description of the problem. """
    return 'Multi-objective Optimisation'

  # Book-keeping ----------------------------------------------------------------
  def _exd_child_update_history(self, qinfo):
    """ Updates to the history specific to optimisation. """
    # Update the best point/val
    # check fidelity
    if self.multi_func_caller.is_mf():
      raise NotImplementedError(_NO_MF_FOR_MOO_ERR_MSG)
    else:
      self._update_opt_point_and_val(qinfo)
    # Now add to history
    self.history.curr_pareto_vals.append(self.curr_pareto_vals)
    self.history.curr_pareto_points.append(self.curr_pareto_points)
    self.history.curr_true_pareto_vals.append(self.curr_true_pareto_vals)
    self.history.curr_true_pareto_points.append(self.curr_true_pareto_points)
    # Any method specific updating
    self._multi_opt_method_update_history(qinfo)

  def _update_opt_point_and_val(self, qinfo, query_is_at_fidel_to_opt=None):
    """ Updates the optimum point and value according the data in qinfo.
        Can be overridden by a child class if you want to do anything differently.
    """
    if query_is_at_fidel_to_opt is not None:
      if not query_is_at_fidel_to_opt:
        # if the fidelity queried at is not fidel_to_opt, then return
        return
    # Optimise curr_opt_val and curr_true_opt_val
    self.curr_pareto_vals, self.curr_pareto_points = update_pareto_set(
      self.curr_pareto_vals, self.curr_pareto_points, qinfo.val, qinfo.point)
    self.curr_true_pareto_vals, self.curr_true_pareto_points = update_pareto_set(
      self.curr_true_pareto_vals, self.curr_true_pareto_points, qinfo.true_val,
      qinfo.point)

  def _multi_opt_method_update_history(self, qinfo):
    """ Any updates to the history specific to the method. """
    pass # Pass by default. Not necessary to override.

  def _get_exd_child_header_str(self):
    """ Header for black box optimisation. """
    ret = '#Pareto=<num_pareto_optimal_points_found>'
    ret += self._get_opt_method_header_str()
    return ret

  @classmethod
  def _get_opt_method_header_str(cls):
    """ Header for optimisation method. """
    return ''

  def _get_exd_child_report_results_str(self):
    """ Returns a string describing the progress in optimisation. """
    best_val_str = '#Pareto: %d'%(len(self.curr_pareto_vals))
    opt_method_str = self._get_opt_method_report_results_str()
    return best_val_str + opt_method_str + ', '

  def _get_opt_method_report_results_str(self):
    """ Any details to include in a child method when reporting results.
        Can be overridden by a child class.
    """
    #pylint: disable=no-self-use
    return ''

  def _exd_child_handle_prev_evals(self):
    """ Handles pre-evaluations. """
    for qinfo in self.options.prev_evaluations.qinfos:
      if self.multi_func_caller.is_mf():
        raise NotImplementedError(_NO_MF_FOR_MOO_ERR_MSG)
      else:
        self._update_opt_point_and_val(qinfo)
      self.prev_eval_points.append(qinfo.point)
      self.prev_eval_vals.append(qinfo.val)
    self.history.prev_eval_points = self.prev_eval_points
    self.history.prev_eval_vals = self.prev_eval_vals

  def _child_run_experiments_initialise(self):
    """ Handles any initialisation before running experiments. """
    self._opt_method_optimise_initalise()

  def _opt_method_optimise_initalise(self):
    """ Any routine to run for a method just before optimisation routine. """
    pass # Pass by default. Not necessary to override.

  def optimise(self, max_capital):
    """ Calling optimise with optimise the function. A wrapper for run_experiments from
        BlackboxExperimenter. """
    ret = self.run_experiments(max_capital)
    return ret

  def _get_final_return_quantities(self):
    """ Return the curr_opt_val, curr_opt_point and history. """
    return self.curr_pareto_vals, self.curr_pareto_points, self.history




"""
  Unit tests for posterior sampling
  -- kvysyara@andrew.cmu.edu
"""
from __future__ import absolute_import
from __future__ import division
import unittest

# pylint: disable=invalid-name
# pylint: disable=no-self-use

import numpy as np

# Local imports
from .continuous import Normal
from .model import Model
from ..utils.base_test_class import BaseTestClass, execute_tests

class PosteriorTestCase(BaseTestClass):
  """ Unit test for posterior sampling with Gaussian likelihood and Gaussian prior"""

  def setUp(self):
    """ Sets up unit tests. """
    self.size = 1000
    self.threshold = 0.01
    self.data = np.random.normal(5, 2, 1000)
    self.m = 0
    self.tau = 4
    self.prior = Normal(self.m, self.tau)
    self.var = 2

  def _check_sample_sizes(self, samples):
    """ Compares the sample sizes with the size parameter"""
    assert self.size == len(samples)

  def _compute_mean(self, samples):
    """ Computes Mean """
    return np.mean(samples)

  def _compute_variance(self, samples):
    """ Computes Variance """
    return np.var(samples)

  def _pdf(self, x):
    """ Computes pdf of model """
    dist = Normal(x, self.var)
    val = self.prior.pdf(x)
    for i in range(len(self.data)):
      val = val*dist.pdf(self.data[i])
    return val

  def _logp(self, x):
    """ Compute log pdf of model """
    dist = Normal(x, self.var)
    val = self.prior.logp(x)
    for i in range(len(self.data)):
      val = val + dist.logp(self.data[i])
    return val

  def _grad_logp(self, x):
    """ Computes gradient of log pdf of model """
    dist = Normal(x, self.var)
    val = self.prior.grad_logp(x)
    for i in range(len(self.data)):
      val = val - dist.grad_logp(self.data[i])
    return val

  @unittest.skip
  def test_posterior_NUTS(self):
    """ Tests posterior estimation using NUTS sampling """
    self.report('Test posterior estimation with Gaussian likelihood and Gaussian prior '
                'using NUTS sampling')

    model_a = Model(self._pdf, self._logp, self._grad_logp)
    samples = model_a.draw_samples('nuts', self.size, np.array([3]), 10)

    avg = 0
    for i in range(len(self.data)):
      avg = avg + self.data[i]
    avg = avg/len(self.data)

    # Map estimation
    val = self.var/len(self.data)
    mean_a = (self.tau*avg)/(self.tau + val) + self.m*(val/(self.tau + val))
    var_a = (val*self.tau)/(self.tau + val)

    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)

    self._check_sample_sizes(samples)
    assert abs(mean_a - mean_r) <= self.threshold
    assert abs(var_a - var_r) <= self.threshold
    self.report('Map Estimation: mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(mean_a, var_a, mean_r, var_r), 'test_result')

  @unittest.skip
  def test_posterior_Slice(self):
    """ Tests posterior estimation using Slice sampling """
    self.report('Test posterior estimation with Gaussian likelihood and Gaussian prior '
                'using Slice sampling')

    model_a = Model(self._pdf, self._logp, self._grad_logp)
    samples = model_a.draw_samples('slice', self.size, np.array([4.5]))

    avg = 0
    for i in range(len(self.data)):
      avg = avg + self.data[i]
    avg = avg/len(self.data)

    # Map estimation
    val = self.var/len(self.data)
    mean_a = (self.tau*avg)/(self.tau + val) + self.m*(val/(self.tau + val))
    var_a = (val*self.tau)/(self.tau + val)

    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)

    self._check_sample_sizes(samples)
    assert abs(mean_a - mean_r) <= self.threshold
    assert abs(var_a - var_r) <= self.threshold
    self.report('Map Estimation: mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(mean_a, var_a, mean_r, var_r), 'test_result')

if __name__ == '__main__':
  execute_tests()
